import logging
import time
from io import StringIO
from typing import Any, Dict, Generator, Optional

import pandas as pd
from pydantic import StrictBool, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.utils.eval.eval_globals import EVAL_GLOBALS

logger = logging.getLogger(__name__)


class EvalExprOpConfigModel(BaseOpConfigModel):
    expr: StrictStr
    metadata_expr: Optional[StrictStr] = None
    msg_var_name: Optional[StrictStr] = None
    time_step_var_name: Optional[StrictStr] = None
    op_manager_var_name: Optional[StrictStr] = None
    log_output: StrictBool = False
    iterate_over_output: StrictBool = False
    msg_data_var_name: Optional[StrictStr] = None
    msg_metadata_var_name: Optional[StrictStr] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class EvalExprOpMetadataModel(BaseOpMetadataModel):
    output_metadata: Optional[Dict[str, Any]]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class EvalExprOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class EvalExprOp(BaseGeneratorOp):
    OP_TYPE = "generator.eval_expr"
    OP_CONFIG_MODEL = EvalExprOpConfigModel
    OP_METADATA_MODEL = EvalExprOpMetadataModel
    OP_AUDIT_MODEL = EvalExprOpAuditModel

    templated_fields = None

    def __init__(
        self,
        expr: str,
        metadata_expr: Optional[str] = None,
        msg_var_name: Optional[str] = None,
        time_step_var_name: Optional[str] = None,
        op_manager_var_name: Optional[str] = None,
        log_output: StrictBool = False,
        iterate_over_output: StrictBool = False,
        msg_data_var_name: Optional[str] = None,
        msg_metadata_var_name: Optional[str] = None,
        **kwargs,
    ):
        self.expr = expr
        self.metadata_expr = metadata_expr
        self.msg_var_name = "msg" if msg_var_name is None else msg_var_name
        self.time_step_var_name = (
            "time_step" if time_step_var_name is None else time_step_var_name
        )
        self.op_manager_var_name = (
            "op_manager" if op_manager_var_name is None else op_manager_var_name
        )
        self.log_output = log_output
        self.iterate_over_output = iterate_over_output
        self.msg_data_var_name = msg_data_var_name
        self.msg_metadata_var_name = msg_metadata_var_name

        self.templated_fields = kwargs.get("templated_fields", [])

        self.op_manager = get_global_op_manager()

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"EvalExprOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=time_step, msg=msg, log_prefix="EvalExprOp.run:")

        logger.info(f"EvalExprOp.run: Evaluating expression: {self.expr=}")

        start = time.time()

        params = {
            **EVAL_GLOBALS,
            self.msg_var_name: msg,
            self.time_step_var_name: time_step,
            self.op_manager_var_name: self.op_manager,
        }

        if self.msg_data_var_name:
            params[self.msg_data_var_name] = msg.data

        if self.msg_metadata_var_name:
            params[self.msg_metadata_var_name] = msg.metadata

        output = eval(self.expr, params)
        end = time.time()

        logger.info(
            f"EvalExprOp.run: Evaluated expression: {type(output)=}. "
            f"Time taken: {end - start:0.3f} seconds"
        )

        # Special logging for pandas DataFrames
        if isinstance(output, pd.DataFrame):
            buf = StringIO()
            output.info(buf=buf)
            buf.seek(0)
            logger.info(f"EvalExprOp.run: DataFrame info:\n{buf.read()}")

        if self.log_output:
            logger.info(f"EvalExprOp.run: Logging output:\n{output=}")

        if self.metadata_expr:
            output_metadata = eval(
                self.metadata_expr,
                {
                    **EVAL_GLOBALS,
                    self.msg_var_name: msg,
                    self.time_step_var_name: time_step,
                    self.op_manager_var_name: self.op_manager,
                },
            )
            logger.info(
                f"EvalExprOp.run: Evaluated metadata expression: "
                f"{type(output_metadata)=}"
            )
            if not isinstance(output_metadata, dict):
                raise ValueError(
                    f"EvalExprOp.run: Expected metadata_expr to evaluate to a dict, "
                    f"but got {type(output_metadata)=}"
                )
        else:
            output_metadata = None

        if self.iterate_over_output:
            logger.info(f"EvalExprOp.run: Iterating over output of {type(output)=}")

            for item_i, item in enumerate(output):
                logger.info(
                    f"EvalExprOp.run: Iterating over output of {type(output)=}, "
                    f"yielding {item_i=} of {type(item)=}"
                )
                yield OpMsg(
                    data=item,
                    metadata=EvalExprOpMetadataModel(output_metadata=output_metadata),
                    audit=EvalExprOpAuditModel(),
                )
        else:
            logger.info(f"EvalExprOp.run: Yielding output of {type(output)=}")
            yield OpMsg(
                data=output,
                metadata=EvalExprOpMetadataModel(output_metadata=output_metadata),
                audit=EvalExprOpAuditModel(),
            )



import logging
from typing import Any, Dict, Generator, List, Optional

import pandas as pd
from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpTimeStepMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseOpConfigModel, BaseTimeStepOp

logger = logging.getLogger(__name__)


class SimpleTimeStepOpConfigModel(BaseOpConfigModel):
    start: StrictStr
    end: StrictStr
    tz: StrictStr
    metadata: Optional[Dict[str, Any]] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SimpleTimeStepOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SimpleTimeStepOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SimpleTimeStepOp(BaseTimeStepOp):
    OP_TYPE = "ts.simple"
    OP_CONFIG_MODEL = SimpleTimeStepOpConfigModel
    OP_METADATA_MODEL = SimpleTimeStepOpMetadataModel
    OP_AUDIT_MODEL = SimpleTimeStepOpAuditModel

    templated_fields = None

    def __init__(
        self,
        start: str,
        end: str,
        tz: str,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        self.start = start
        self.end = end
        self.tz = tz
        self.metadata = metadata

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(self) -> Generator[OpTimeStepMsg, None, None]:
        """."""
        logger.info(f"SimpleTimeStepOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=None, msg=None, log_prefix="SimpleTimeStepOp.run:")

        time_step = TimeStep(
            start=self.start,
            end=self.end,
            tz=self.tz,
            metadata=self.metadata,
        )

        logger.info(f"SimpleTimeStepOp.run: {time_step=}")

        yield OpTimeStepMsg(
            data=time_step,
            metadata=SimpleTimeStepOpMetadataModel(),
            audit=SimpleTimeStepOpAuditModel(),
        )



import io
import json
import logging
import time
import uuid
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple, Union

import pandas as pd
import requests
from pydantic import StrictBool, StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp

logger = logging.getLogger(__name__)


class PriceindxDownloadExportPipelineOpConfigModel(BaseOpConfigModel):
    http_conn_id: StrictStr
    setup_id: StrictStr
    setup_name: StrictStr
    # Could be templated
    export_type: Union[StrictStr, Literal["all_offers_excluded", "all_offers_included"]]
    accepted_status_codes: Optional[conlist(int, min_items=1)] = None
    max_wait_time_for_export: int = 900
    sleep_time_between_requests: int = 5
    timeout_seconds_per_request: float = 180

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PriceindxDownloadExportPipelineOpMetadataModel(BaseOpMetadataModel):
    setup_id: str
    setup_name: str
    export_type: str
    report_name: str
    export_id: int

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PriceindxDownloadExportPipelineOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PriceindxDownloadExportPipelineOp(BaseGeneratorOp):
    """.

    This is an operator to download PriceIndx exports which are available via the
    webpage. This is a full pipeline that does the following:
    * Logs into the website
    * Changes the customer setup i.e. change the setup_id we want to download the
      export for different setups or countries etc.
    * Starts an export
    * Waits for the export to be ready
    * Downloads the export file
    * Deletes the export

    This is meant to be a temporary Op while we wait for PriceIndx to implement an
    API to access this data.

    IMPORTANT:
    1. We hardcode URLs
    2. We hardcode the JSON payload for the POST request that starts the export.
       This is the most likely thing to CHANGE or BREAK in the future.
       In order to update the JSON payload, you need to goto the website and start an
       export manually and note the JSON payload using the browser's developer tools.
    """

    OP_TYPE = "generator.priceindx.download_export_pipeline"
    OP_CONFIG_MODEL = PriceindxDownloadExportPipelineOpConfigModel
    OP_METADATA_MODEL = PriceindxDownloadExportPipelineOpMetadataModel
    OP_AUDIT_MODEL = PriceindxDownloadExportPipelineOpAuditModel

    templated_fields = None

    def __init__(
        self,
        http_conn_id: str,
        setup_id: str,
        setup_name: str,
        export_type: Literal["all_offers_excluded", "all_offers_included"],
        accepted_status_codes: Optional[List[int]] = None,
        max_wait_time_for_export: int = 900,
        sleep_time_between_requests: int = 5,
        timeout_seconds_per_request: float = 180,
        **kwargs,
    ):
        """."""
        self.http_conn_id = http_conn_id
        self.setup_id = setup_id
        self.setup_name = setup_name
        self.export_type = export_type
        self.max_wait_time_for_export = max_wait_time_for_export
        self.sleep_time_between_requests = sleep_time_between_requests
        self.accepted_status_codes = (
            accepted_status_codes if accepted_status_codes else [200]
        )
        self.timeout_seconds_per_request = timeout_seconds_per_request

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(op_id=self.http_conn_id)

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"PriceindxDownloadExportPipelineOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="PriceindxDownloadExportPipelineOp.run:",
        )

        session = requests.session()
        session, cookie_jar, response, csrf_token = self.first_communication(session)
        time.sleep(self.sleep_time_between_requests)
        session, cookie_jar, response, csrf_token = self.login(
            session,
            csrf_token,
            username=self.http_conn.username,
            password=self.http_conn.password,
        )
        time.sleep(self.sleep_time_between_requests)
        session, cookie_jar, response, csrf_token = self.change_customersetup(
            session, csrf_token
        )
        time.sleep(self.sleep_time_between_requests)
        if self.export_type == "all_offers_included":
            post_export_method = self.post_export_large_file
        elif self.export_type == "all_offers_excluded":
            post_export_method = self.post_export_small_file
        else:
            raise NotImplementedError()
        (
            session,
            cookie_jar,
            response,
            csrf_token,
            report_name,
        ) = post_export_method(session, csrf_token)
        time.sleep(self.sleep_time_between_requests)
        (
            session,
            cookie_jar,
            response,
            csrf_token,
            export_id,
        ) = self.wait_until_export_is_ready(
            session, csrf_token, report_name=report_name
        )
        time.sleep(self.sleep_time_between_requests)
        (
            session,
            cookie_jar,
            response,
            csrf_token,
            bytesio,
            etl_metadata_json,
        ) = self.download_export_file(session, csrf_token, export_id)
        time.sleep(self.sleep_time_between_requests)
        session, cookie_jar, response, csrf_token = self.delete_export(
            session, csrf_token, export_id
        )

        df = pd.read_csv(bytesio, sep=",")
        records = df.to_dict(orient="records")
        records = [
            {
                "data": record,
                "etl_metadata": etl_metadata_json,
            }
            for record in records
        ]

        yield OpMsg(
            data=records,
            metadata=PriceindxDownloadExportPipelineOpMetadataModel(
                setup_id=self.setup_id,
                setup_name=self.setup_name,
                export_type=self.export_type,
                report_name=report_name,
                export_id=export_id,
            ),
            audit=PriceindxDownloadExportPipelineOpAuditModel(),
        )

    def first_communication(self, session):
        """."""
        logger.info(f"PriceindxDownloadExportPipelineOp.first_communication: Running.")
        response = session.get("https://customer.priceindx.com/accounts/login/")
        csrf_token = response.cookies["csrftoken"]
        cookie_jar = session.cookies
        if response.status_code not in self.accepted_status_codes:
            raise ValueError(
                f"PriceindxDownloadExportPipelineOp.first_communication: ",
                f"{response.status_code=} not in self.accepted_status_codes. "
                f"{response.text=}",
            )
        logger.info(f"PriceindxDownloadExportPipelineOp.first_communication: Done.")
        return session, cookie_jar, response, csrf_token

    def login(self, session, csrf_token, username, password):
        logger.info(f"PriceindxDownloadExportPipelineOp.login: Running.")
        headers = {"X-CSRFToken": csrf_token}
        data = {
            "username": username,
            "password": password,
            "csrfmiddlewaretoken": csrf_token,
        }
        response = session.post(
            "https://customer.priceindx.com/accounts/login/",
            data=data,
            headers=headers,
        )
        csrf_token = response.cookies["csrftoken"]
        cookie_jar = session.cookies
        if response.status_code not in self.accepted_status_codes:
            raise ValueError(
                f"PriceindxDownloadExportPipelineOp.login: ",
                f"{response.status_code=} not in self.accepted_status_codes. "
                f"{response.text=}",
            )
        logger.info(f"PriceindxDownloadExportPipelineOp.login: Done.")
        return session, cookie_jar, response, csrf_token

    def change_customersetup(self, session, csrf_token):
        logger.info(
            f"PriceindxDownloadExportPipelineOp.change_customersetup: Running "
            f"for {self.setup_id=} and {self.setup_name=}."
        )
        data = {"csrfmiddlewaretoken": csrf_token, "setupid": self.setup_id}
        headers = {"X-CSRFToken": csrf_token}
        response = session.post(
            "https://customer.priceindx.com/account/ajax/change_customersetup/",
            data=data,
            headers=headers,
        )
        cookie_jar = session.cookies
        if response.status_code not in self.accepted_status_codes:
            raise ValueError(
                f"PriceindxDownloadExportPipelineOp.change_customersetup: ",
                f"{response.status_code=} not in self.accepted_status_codes. "
                f"{response.text=}",
            )
        logger.info(f"PriceindxDownloadExportPipelineOp.change_customersetup: Done.")
        return session, cookie_jar, response, csrf_token

    def post_export_large_file(self, session, csrf_token):
        report_name = str(uuid.uuid4())
        logger.info(
            f"PriceindxDownloadExportPipelineOp.post_export_large_file: Running for "
            f"{report_name=}."
        )
        data = json.loads(
            """
            {"shared":false,"cfg":{"col_order":{"product":{"columns":[{"label":"Image","key":"image","vis":true},{"label":"SKU","key":"sku","vis":true},{"label":"Category","key":"category","vis":true},{"label":"Brand","key":"brand","vis":true},{"label":"Name","key":"name","vis":true},{"label":"Position","key":"position","vis":true},{"label":"Price pos.","key":"price_pos","vis":true},{"label":"Price pos. change","key":"price_pos_change","vis":true},{"label":"Price","key":"price","vis":true},{"label":"Price change","key":"price_change","vis":true},{"label":"Purchase Price","key":"in_price","vis":true},{"label":"Margin","key":"margin","vis":true},{"label":"Profit","key":"profit","vis":true},{"label":"Stock","key":"stock","vis":true}]},"dp":{"columns":[{"label":"Suggested price","key":"suggested_price","vis":true},{"label":"Price Margin","key":"suggested_price_margin","vis":true},{"label":"Price Based On","key":"suggestion_based_on","vis":true},{"label":"Group / Rule / Limit","key":"group","vis":true},{"label":"Limit: Lower margin","key":"margin_lower_limit","vis":true},{"label":"Limit: Upper margin","key":"margin_upper_limit","vis":true},{"label":"Price clamped","key":"clamped","vis":true}],"groupsets":[{"label":"Group set 0","key":"0","vis":true}]},"competitors":{"columns":[{"label":"Lowest competitor","key":"lowest_comp","vis":true}]}},"date_settings":{"report_sts":1692655200000,"ga_timestamps":{},"changes_occur":"last_week","offer_dates":["2023-08-22"]},"table_filters":{"dp_group_set":["0"]},"active_sorting":[],"global_filters":{},"nulls_last_sort":false},"name":"EMPTY","offers_included":true}
            """
        )
        data["name"] = report_name
        headers = {"X-CSRFToken": csrf_token}
        response = session.post(
            "https://customer.priceindx.com/reports/ajax_rundeck_table_export/",
            headers=headers,
            json=data,
        )
        cookie_jar = session.cookies
        if response.status_code not in self.accepted_status_codes:
            raise ValueError(
                f"PriceindxDownloadExportPipelineOp.post_export_large_file: ",
                f"{response.status_code=} not in self.accepted_status_codes. "
                f"{response.text=}",
            )
        logger.info(
            f"PriceindxDownloadExportPipelineOp.post_export_large_file: Started export "
            f"for {report_name=}."
        )
        return session, cookie_jar, response, csrf_token, report_name

    def post_export_small_file(self, session, csrf_token):
        report_name = str(uuid.uuid4())
        logger.info(
            f"PriceindxDownloadExportPipelineOp.post_export_small_file: Running for "
            f"{report_name=}."
        )
        data = json.loads(
            """
            {"shared":false,"cfg":{"col_order":{"product":{"columns":[{"label":"Image","key":"image","vis":true},{"label":"SKU","key":"sku","vis":true},{"label":"Category","key":"category","vis":true},{"label":"Brand","key":"brand","vis":true},{"label":"Name","key":"name","vis":true},{"label":"Position","key":"position","vis":true},{"label":"Price pos.","key":"price_pos","vis":true},{"label":"Price pos. change","key":"price_pos_change","vis":true},{"label":"Price","key":"price","vis":true},{"label":"Price change","key":"price_change","vis":true},{"label":"Purchase Price","key":"in_price","vis":true},{"label":"Margin","key":"margin","vis":true},{"label":"Profit","key":"profit","vis":true},{"label":"Stock","key":"stock","vis":true}]},"dp":{"columns":[{"label":"Suggested price","key":"suggested_price","vis":true},{"label":"Price Margin","key":"suggested_price_margin","vis":true},{"label":"Price Based On","key":"suggestion_based_on","vis":true},{"label":"Group / Rule / Limit","key":"group","vis":true},{"label":"Limit: Lower margin","key":"margin_lower_limit","vis":true},{"label":"Limit: Upper margin","key":"margin_upper_limit","vis":true},{"label":"Price clamped","key":"clamped","vis":true}],"groupsets":[{"label":"Group set 0","key":"0","vis":true}]},"competitors":{"columns":[{"label":"Lowest competitor","key":"lowest_comp","vis":true}]}},"date_settings":{"report_sts":1692655200000,"ga_timestamps":{},"changes_occur":"last_week","offer_dates":["2023-08-22"]},"table_filters":{"dp_group_set":["0"]},"active_sorting":[],"global_filters":{},"nulls_last_sort":false},"name":"EMPTY","offers_included":false}
            """
        )
        data["name"] = report_name
        headers = {"X-CSRFToken": csrf_token}
        response = session.post(
            "https://customer.priceindx.com/reports/ajax_rundeck_table_export/",
            headers=headers,
            json=data,
        )
        cookie_jar = session.cookies
        if response.status_code not in self.accepted_status_codes:
            raise ValueError(
                f"PriceindxDownloadExportPipelineOp.post_export_small_file: ",
                f"{response.status_code=} not in self.accepted_status_codes. "
                f"{response.text=}",
            )
        logger.info(
            f"PriceindxDownloadExportPipelineOp.post_export_small_file: Started export "
            f"for {report_name=}."
        )
        return session, cookie_jar, response, csrf_token, report_name

    def get_exports(self, session, csrf_token):
        logger.info(f"PriceindxDownloadExportPipelineOp.get_exports: Running.")
        headers = {"X-CSRFToken": csrf_token}
        response = session.get(
            "https://customer.priceindx.com/reports/get_user_exports/",
            headers=headers,
        )
        cookie_jar = session.cookies
        if response.status_code not in self.accepted_status_codes:
            raise ValueError(
                f"PriceindxDownloadExportPipelineOp.get_exports: ",
                f"{response.status_code=} not in self.accepted_status_codes. "
                f"{response.text=}",
            )
        logger.info(f"PriceindxDownloadExportPipelineOp.get_exports: Done.")
        return session, cookie_jar, response, csrf_token

    def wait_until_export_is_ready(
        self,
        session,
        csrf_token,
        report_name,
    ):
        logger.info(
            f"PriceindxDownloadExportPipelineOp.wait_until_export_is_ready: Running."
        )
        start_time = time.time()
        while True:
            session, cookie_jar, response, csrf_token = self.get_exports(
                session, csrf_token
            )
            if response.status_code not in self.accepted_status_codes:
                raise ValueError(
                    f"PriceindxDownloadExportPipelineOp.wait_until_export_is_ready: ",
                    f"{response.status_code=} not in self.accepted_status_codes",
                )

            exports = response.json()["exports"]
            for export in exports:
                if export["name"] == report_name:
                    if export["status"] == 1:
                        logger.info(
                            f"PriceindxDownloadExportPipelineOp.wait_until_export_is_ready: "
                            f"Done waiting for export {report_name=} to be ready. "
                            f"Export is ready with export_id={export['id']}."
                        )
                        return (
                            session,
                            cookie_jar,
                            response,
                            csrf_token,
                            export["id"],
                        )
                    else:
                        logger.info(
                            f"PriceindxDownloadExportPipelineOp.wait_until_export_is_ready: "
                            f"Waiting for export {report_name=} to be ready. "
                            f"Sleeping for {self.sleep_time_between_requests} seconds."
                        )
                        break
            else:
                raise ValueError(
                    f"PriceindxDownloadExportPipelineOp.wait_until_export_is_ready: "
                    f"Export with {report_name=} not found in exports."
                )

            time.sleep(self.sleep_time_between_requests)
            wait_time = time.time() - start_time

            if wait_time > self.max_wait_time_for_export:
                raise ValueError(
                    f"PriceindxDownloadExportPipelineOp.wait_until_export_is_ready: "
                    f"Exceeded {self.max_wait_time_for_export=} while waiting for "
                    f"export with {report_name=}."
                )

    def download_export_file(self, session, csrf_token, export_id):
        logger.info(f"PriceindxDownloadExportPipelineOp.download_export_file: Running.")
        headers = {"X-CSRFToken": csrf_token}
        etl_request_start_ts = pd.Timestamp.now(tz="UTC").isoformat()
        response = session.get(
            f"https://customer.priceindx.com/reports/get_file/{export_id}",
            headers=headers,
        )
        etl_response_end_ts = pd.Timestamp.now(tz="UTC").isoformat()
        etl_metadata_json = {
            "etl_request_start_ts": etl_request_start_ts,
            "etl_response_end_ts": etl_response_end_ts,
        }
        cookie_jar = session.cookies
        if response.status_code not in self.accepted_status_codes:
            raise ValueError(
                f"PriceindxDownloadExportPipelineOp.download_export_file: ",
                f"{response.status_code=} not in self.accepted_status_codes. "
                f"{response.text=}",
            )
        bytesio = io.BytesIO(response.content)
        bytesio.seek(0)
        logger.info(
            f"PriceindxDownloadExportPipelineOp.download_export_file: Done "
            f"downloading export file with {export_id=}. "
            f"File size approx {bytesio.getbuffer().nbytes//1000:,} KiB"
        )
        return session, cookie_jar, response, csrf_token, bytesio, etl_metadata_json

    def delete_export(self, session, csrf_token, export_id):
        logger.info(f"PriceindxDownloadExportPipelineOp.delete_export: Running.")
        data = {"id": export_id}
        headers = {"X-CSRFToken": csrf_token}
        response = session.post(
            f"https://customer.priceindx.com/reports/delete_user_export/",
            headers=headers,
            json=data,
        )
        cookie_jar = session.cookies
        if response.status_code not in self.accepted_status_codes:
            raise ValueError(
                f"PriceindxDownloadExportPipelineOp.delete_export: ",
                f"{response.status_code=} not in self.accepted_status_codes. "
                f"{response.text=}",
            )
        logger.info(
            f"PriceindxDownloadExportPipelineOp.delete_export: Done "
            f"deleting export file with {export_id=}"
        )
        return session, cookie_jar, response, csrf_token



import logging
from typing import Literal, Union

from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.list import ListConnOp

logger = logging.getLogger(__name__)


class ListPutConsumerOpConfigModel(BaseOpConfigModel):
    list_conn_id: StrictStr
    put_type: Literal["append", "extend"]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class ListPutConsumerOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class ListPutConsumerOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class ListPutConsumerOp(BaseConsumerOp):
    OP_TYPE = "consumer.list.put"
    OP_CONFIG_MODEL = ListPutConsumerOpConfigModel
    OP_METADATA_MODEL = ListPutConsumerOpMetadataModel
    OP_AUDIT_MODEL = ListPutConsumerOpAuditModel

    templated_fields = None

    def __init__(
        self, list_conn_id: str, put_type: Literal["append", "extend"], **kwargs
    ):
        self.list_conn_id = list_conn_id
        self.put_type = put_type

        op_manager = get_global_op_manager()

        self.list_conn: Union[ListConnOp] = op_manager.get_connector(
            op_id=self.list_conn_id
        )

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(self, time_step: TimeStep, msg: OpMsg) -> OpMsg:
        """."""
        logger.info(f"ListPutConsumerOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="ListPutConsumerOp.run:"
        )

        logger.info(
            f"ListPutConsumerOp.run: Putting item of {type(msg.data)=} to list "
            f"with {self.put_type=}. Current {self.list_conn.size()=}"
        )

        if self.put_type == "append":
            self.list_conn.put_append(msg.data)
        elif self.put_type == "extend":
            self.list_conn.put_extend(msg.data)
        else:
            raise NotImplementedError()

        logger.info(
            f"ListPutConsumerOp.run: Done Putting item of {type(msg.data)=} to list "
            f"with {self.put_type=}. New {self.list_conn.size()=}"
        )

        return OpMsg(
            data=None,
            metadata=ListPutConsumerOpMetadataModel(),
            audit=ListPutConsumerOpAuditModel(),
        )



import sqlalchemy.dialects.postgresql


def lookup_sqlalchemy_dtype(dtype: str):
    """Get dtype from full str."""
    if dtype.startswith("sqlalchemy.dialects.postgresql"):
        return getattr(sqlalchemy.dialects.postgresql, dtype.split(".")[-1])
    else:
        raise NotImplementedError()



import logging
from typing import Any, Dict, Generator, List, Optional

import pandas as pd
from pydantic import BaseModel, StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp

logger = logging.getLogger(__name__)


class BladeGetTokenOpConfigModel(BaseOpConfigModel):
    http_conn_id: StrictStr
    accepted_status_codes: Optional[conlist(int, min_items=1)] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeGetTokenOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeGetTokenOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeGetTokenResponseDataModel(BaseModel):
    session_token: StrictStr
    expiry: StrictStr


class BladeGetTokenResponseModel(BaseModel):
    meta: Dict[str, Any]
    data: BladeGetTokenResponseDataModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeGetTokenOp(BaseGeneratorOp):
    OP_TYPE = "generator.blade.get_token"
    OP_CONFIG_MODEL = BladeGetTokenOpConfigModel
    OP_METADATA_MODEL = BladeGetTokenOpMetadataModel
    OP_AUDIT_MODEL = BladeGetTokenOpAuditModel

    templated_fields = None

    def __init__(
        self,
        http_conn_id: str,
        accepted_status_codes: Optional[List[int]] = None,
        **kwargs,
    ):
        """."""
        self.http_conn_id = http_conn_id
        self.accepted_status_codes = accepted_status_codes

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(op_id=self.http_conn_id)

        # Loaded in run
        self.auth_model: Optional[BladeGetTokenResponseModel] = None

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"BladeGetTokenOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="BladeGetTokenOp.run:"
        )

        logger.info(
            f"BladeGetTokenOp.run: Getting auth token from {self.http_conn_id=}"
        )

        # Call the refresh token the first time, so that the self.auth_model is set
        # You can then call self.token to get the token.
        self.refresh_token()

        logger.info(
            f"BladeGetTokenOp.run: Successfully fetched auth_token! The attribute "
            f"`self.auth_model` has been set for {self.http_conn_id=}. "
            f"The (refreshed) token is now accessible by calling "
            f"self.token OR op.token OR msg.data.token"
        )

        yield OpMsg(
            data=self,
            metadata=BladeGetTokenOpMetadataModel(),
            audit=BladeGetTokenOpAuditModel(),
        )

    ####################################################################################
    # PROPERTIES & METHODS
    ####################################################################################
    def refresh_token(self):
        """."""
        status_code, auth_json = self.http_conn.call(
            method="post",
            url=f"{self.http_conn.base_url}/v1/auth/login",
            requests_kwargs={
                "headers": {"Content-Type": "application/json"},
                "json": {
                    "username": self.http_conn.username,
                    "password": self.http_conn.password,
                },
            },
            accepted_status_codes=self.accepted_status_codes,
            return_type="json",
        )

        self.auth_model = BladeGetTokenResponseModel(**auth_json)

    @property
    def token(self) -> str:
        """."""
        if self.auth_model is None:
            raise ValueError(
                "BladeGetTokenOp.token: self.auth_model is not set, "
                "please call op.run() first!"
            )

        expiry = pd.Timestamp(self.auth_model.data.expiry, tz="UTC")
        now = pd.Timestamp.now(tz="UTC")
        now_plus_315 = now + pd.Timedelta(seconds=315)

        if expiry < now_plus_315:
            logger.info(
                f"BladeGetTokenOp.token: Token has either expired or will expire in "
                f"315 seconds. Refreshing token. {expiry=}, {now=}"
            )
            self.refresh_token()
        else:
            logger.info(
                f"BladeGetTokenOp.token: Token has not yet expired. It will expire in "
                f"{(expiry - now).seconds} seconds. {expiry=}, {now=}"
            )

        return self.auth_model.data.session_token



import base64
import logging
import traceback
from typing import Any

import ruamel.yaml

from .main import core
from .messages.op_msg import OpMsg

logger = logging.getLogger(__name__)


def run_lambda(b64_config_yaml: str, root_data: Any = None):
    """Entrypoint for the nr_ops package."""
    msg = (
        "Running nr_ops.main.run(), no logging initialized yet. Attempting to "
        "configure default logging..."
    )
    print(msg)
    logger.info(msg)

    # Configure default logging until the user provided logging config is loaded.
    # dictConfig(config=LoggingDictConfigModel().dict())
    logger.info("Configured default logging!")

    try:
        config_str = base64.b64decode(b64_config_yaml)
        config = ruamel.yaml.YAML(typ="safe", pure=True).load(config_str)
        root_msg = OpMsg(data=root_data, metadata={}, audit={})
        yield from core(config=config, return_generator=True, root_msg=root_msg)
    except Exception:
        logger.exception(
            f"Exception occurred in nr_ops.main.core()!\n{traceback.format_exc()}"
        )
        raise

    logger.info("Done running nr_ops.main.run_lambda!")



import itertools
import logging
from typing import Generator, Optional

from pydantic import StrictInt

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class BatchGeneratorOpConfigModel(BaseOpConfigModel):
    batch_size: StrictInt

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BatchGeneratorOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BatchGeneratorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BatchGeneratorOp(BaseGeneratorOp):
    OP_TYPE = "generator.batch"
    OP_CONFIG_MODEL = BatchGeneratorOpConfigModel
    OP_METADATA_MODEL = BatchGeneratorOpMetadataModel
    OP_AUDIT_MODEL = BatchGeneratorOpAuditModel

    templated_fields = None

    def __init__(
        self,
        batch_size: int,
        **kwargs,
    ):
        """."""
        self.batch_size = batch_size

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"BatchGeneratorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="BatchGeneratorOp.run:"
        )

        # ref: https://stackoverflow.com/a/24527424
        def batcher(iterable, batch_size=10):
            iterator = iter(iterable)
            for first in iterator:
                yield itertools.chain(
                    [first], itertools.islice(iterator, batch_size - 1)
                )

        for batch_i, batch in enumerate(batcher(msg.data, batch_size=self.batch_size)):
            logger.info(
                f"BatchGeneratorOp: Yielding batch {batch_i + 1} (1-index) "
                f"of size {self.batch_size=}."
            )
            yield OpMsg(
                data=batch,
                metadata=BatchGeneratorOpMetadataModel(),
                audit=BatchGeneratorOpAuditModel(),
            )



import json
import logging
import time
import traceback
from concurrent.futures import as_completed
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple

import pandas as pd
import requests
from pydantic import Field, StrictBool, StrictInt, StrictStr, conlist, validator
from requests_futures.sessions import FuturesSession

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp

logger = logging.getLogger(__name__)


class BrazeGenericPostJsonDataAsyncOpConfigModel(BaseOpConfigModel):
    http_conn_id: StrictStr
    endpoint: StrictStr
    accepted_status_codes: Optional[conlist(int, min_items=1)] = Field(
        default_factory=lambda: [200, 201]
    )
    number_of_concurrent_workers: StrictInt = 1
    timeout_seconds_per_request: StrictInt = 60

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    @validator("endpoint", pre=False)
    def validate_endpoint(cls, endpoint: str):
        if not endpoint.startswith("/"):
            raise ValueError(f"{endpoint=} must start with a forward slash (/). ")

        return endpoint


class BrazeGenericPostJsonDataAsyncOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BrazeGenericPostJsonDataAsyncOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BrazeGenericPostJsonDataAsyncOp(BaseConsumerOp):
    """."""

    OP_TYPE = "consumer.braze.generic_post_json_data_async"
    OP_CONFIG_MODEL = BrazeGenericPostJsonDataAsyncOpConfigModel
    OP_METADATA_MODEL = BrazeGenericPostJsonDataAsyncOpMetadataModel
    OP_AUDIT_MODEL = BrazeGenericPostJsonDataAsyncOpAuditModel

    templated_fields = None

    def __init__(
        self,
        http_conn_id: StrictStr,
        endpoint: StrictStr,
        accepted_status_codes: Optional[List[int]] = None,
        timeout_seconds_per_request: int = 60,
        number_of_concurrent_workers: int = 1,
        **kwargs,
    ):
        """."""
        self.http_conn_id = http_conn_id
        self.endpoint = endpoint
        self.accepted_status_codes = (
            accepted_status_codes if accepted_status_codes else [200]
        )
        self.timeout_seconds_per_request = timeout_seconds_per_request
        self.number_of_concurrent_workers = number_of_concurrent_workers

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(op_id=self.http_conn_id)

    def run(self, time_step: TimeStep, msg: Optional[OpMsg] = None) -> OpMsg:
        """."""
        logger.info(f"BrazeGenericPostJsonDataAsyncOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="BrazeGenericPostJsonDataAsyncOp.run:",
        )

        if not isinstance(msg.data, list):
            raise TypeError(
                f"BrazeGenericPostJsonDataAsyncOp: msg.data is expected to be a "
                f"list/batch of json payloads. Instead received {type(msg.data)=}."
            )

        for payload_index, payload in enumerate(msg.data):
            if not isinstance(payload, dict):
                raise TypeError(
                    f"BrazeGenericPostJsonDataAsyncOp: "
                    f"msg.data[{payload_index}] is expected to be a dict/json payload. "
                    f"Instead received {type(payload)=}."
                )

        logger.info(
            f"BrazeGenericPostJsonDataAsyncOp.run: Initializing FuturesSession "
            f"with {self.number_of_concurrent_workers=}."
        )

        url = self.http_conn.base_url + self.endpoint
        logger.info(
            f"BrazeGenericPostJsonDataAsyncOp: Creating {len(msg.data)=} "
            f"requests futures with {self.timeout_seconds_per_request=} and {url=}."
        )
        futures = []

        session = FuturesSession(
            max_workers=self.number_of_concurrent_workers,
            adapter_kwargs={
                "pool_connections": self.number_of_concurrent_workers,
                "pool_maxsize": self.number_of_concurrent_workers,
            },
        )

        for payload_index, payload in enumerate(msg.data):
            # print(json.dumps(payload, indent=2))
            future = session.post(
                url=url,
                headers={
                    **self.http_conn.extras,
                },
                json=payload,
                timeout=self.timeout_seconds_per_request,
            )
            future.payload_index = payload_index
            future.request_start = pd.Timestamp.now(tz="UTC").isoformat()
            futures.append(future)

        logger.info(
            f"BrazeGenericPostJsonDataAsyncOp.run: " f"Done creating requests futures. "
        )

        failed_responses = {}
        successful_responses = {}
        wall_time_start = time.time()

        for future in as_completed(futures):
            try:
                response = future.result()
            except requests.exceptions.RequestException:
                logger.info(
                    f"BrazeGenericPostJsonDataAsyncOp: Caught RequestException "
                    f"for {future.payload_index=}. \n{traceback.format_exc()}\n"
                    f"CONTINUING WITH OTHER FUTURE REQUESTS..."
                )
                failed_responses[future.payload_index] = {
                    "payload_index": future.payload_index,
                    "exception": traceback.format_exc(),
                }
                continue

            # For failed responses, we will log the status code and the response text.
            if response.status_code not in self.accepted_status_codes:
                failed_responses[future.payload_index] = {
                    "payload_index": future.payload_index,
                    "status_code": response.status_code,
                    "text": response.text,
                    "response_time_taken_in_seconds": response.elapsed.total_seconds(),
                }
            # For successful responses, we will log the status code.
            else:
                successful_responses[future.payload_index] = {
                    "status_code": response.status_code,
                    "response_time_taken_in_seconds": response.elapsed.total_seconds(),
                }

        wall_time_end = time.time()

        if successful_responses:
            logger.info(
                f"BrazeGenericPostJsonDataAsyncOp.run: "
                f"SUCCESSFUL RESPONSES identified by their payload indexes "
                f"have acceptable status codes:\n"
                f"{json.dumps(successful_responses, indent=2, sort_keys=True)}"
            )

        if failed_responses:
            raise ValueError(
                f"BrazeGenericPostJsonDataAsyncOp.run: "
                f"FAILED RESPONSES identified by their payload indexes do not "
                f"have acceptable status codes:\n"
                f"{json.dumps(failed_responses, indent=2, sort_keys=True)}"
            )

        df_successful_response_times = pd.DataFrame(
            data={
                "successful_response_times_in_seconds": [
                    response["response_time_taken_in_seconds"]
                    for response in successful_responses.values()
                ],
            }
        )

        logger.info(
            f"BrazeGenericPostJsonDataAsyncOp.run: \n"
            f"{df_successful_response_times.describe(percentiles=[0.25, 0.5, 0.75, 0.9, 0.95])}\n\n"
            f"successful_response_times_total = {df_successful_response_times.successful_response_times_in_seconds.sum()}\n"
            f"wall_time = {wall_time_end - wall_time_start}\n"
        )

        df_failed_response_times = pd.DataFrame(
            data={
                "failed_response_times_in_seconds": [
                    response["response_time_taken_in_seconds"]
                    for response in failed_responses.values()
                ],
            }
        )

        logger.info(
            f"BrazeGenericPostJsonDataAsyncOp.run: \n"
            f"{df_failed_response_times.describe(percentiles=[0.25, 0.5, 0.75, 0.9, 0.95])}\n\n"
            f"failed_response_times_total = {df_failed_response_times.failed_response_times_in_seconds.sum()}\n"
            f"wall_time = {wall_time_end - wall_time_start}\n"
        )

        return OpMsg(
            data=None,
            metadata=BrazeGenericPostJsonDataAsyncOpMetadataModel(),
            audit=BrazeGenericPostJsonDataAsyncOpAuditModel(),
        )



import logging
from io import StringIO
from typing import Any, Dict, Generator, Optional

import pandas as pd
from pydantic import StrictBool, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.utils.eval.eval_globals import EVAL_GLOBALS

logger = logging.getLogger(__name__)


class EvalExprAsMetadataOpConfigModel(BaseOpConfigModel):
    expr: StrictStr
    msg_var_name: Optional[StrictStr] = None
    time_step_var_name: Optional[StrictStr] = None
    op_manager_var_name: Optional[StrictStr] = None
    log_output: StrictBool = False

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class EvalExprAsMetadataOpMetadataModel(BaseOpMetadataModel):
    output_metadata: Any

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class EvalExprAsMetadataOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class EvalExprAsMetadataOp(BaseGeneratorOp):
    """Store output of eval_expr as metadata. Return input data as is."""

    OP_TYPE = "generator.eval_expr_as_metadata"
    OP_CONFIG_MODEL = EvalExprAsMetadataOpConfigModel
    OP_METADATA_MODEL = EvalExprAsMetadataOpMetadataModel
    OP_AUDIT_MODEL = EvalExprAsMetadataOpAuditModel

    templated_fields = None

    def __init__(
        self,
        expr: str,
        msg_var_name: Optional[str] = None,
        time_step_var_name: Optional[str] = None,
        op_manager_var_name: Optional[str] = None,
        log_output: StrictBool = False,
        **kwargs,
    ):
        self.expr = expr
        self.msg_var_name = "msg" if msg_var_name is None else msg_var_name
        self.time_step_var_name = (
            "time_step" if time_step_var_name is None else time_step_var_name
        )
        self.op_manager_var_name = (
            "op_manager" if op_manager_var_name is None else op_manager_var_name
        )
        self.log_output = log_output

        self.templated_fields = kwargs.get("templated_fields", [])

        self.op_manager = get_global_op_manager()

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"EvalExprAsMetadataOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="EvalExprAsMetadataOp.run:"
        )

        output = eval(
            self.expr,
            {
                **EVAL_GLOBALS,
                self.msg_var_name: msg,
                self.time_step_var_name: time_step,
                self.op_manager_var_name: self.op_manager,
            },
        )

        logger.info(f"EvalExprAsMetadataOp.run: Evaluated expression: {type(output)=}")

        # Special logging for pandas DataFrames
        if isinstance(output, pd.DataFrame):
            buf = StringIO()
            output.info(buf=buf)
            buf.seek(0)
            logger.info(f"EvalExprAsMetadataOp.run: DataFrame info:\n{buf.read()}")

        if self.log_output:
            logger.info(f"EvalExprAsMetadataOp.run: Logging output:\n{output=}")

        logger.info(
            f"EvalExprAsMetadataOp.run: STORING output of {type(output)=} as metadata "
            f"of this op. Yielding input data of this op as output."
        )

        # Yield input data as output.
        # Store output as metadata.
        yield OpMsg(
            data=msg.data,
            metadata=EvalExprAsMetadataOpMetadataModel(output_metadata=output),
            audit=EvalExprAsMetadataOpAuditModel(),
        )



import logging
import os
from typing import Any

from google.ads.googleads.client import GoogleAdsClient
from pydantic import BaseModel, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class GoogleAdsConnectorOpConfigModel(BaseOpConfigModel):
    api_version: StrictStr = "v15"

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAdsConnectorOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAdsConnectorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAdsConnectorOp(BaseConnectorOp):
    OP_TYPE = "connector.google_ads"
    OP_CONFIG_MODEL = GoogleAdsConnectorOpConfigModel
    OP_METADATA_MODEL = GoogleAdsConnectorOpMetadataModel
    OP_AUDIT_MODEL = GoogleAdsConnectorOpAuditModel

    templated_fields = None

    def __init__(self, api_version: str = "v15", **kwargs):
        self.api_version = api_version

        self.templated_fields = kwargs.get("templated_fields", [])

        self.client: Any = None

    def run(self) -> OpMsg:
        """."""
        logger.info(f"GoogleAdsConnectorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="GoogleAdsConnectorOp.run:"
        )

        if "GOOGLE_ADS_CONFIGURATION_FILE_PATH" not in os.environ:
            raise ValueError(
                f"GOOGLE_ADS_CONFIGURATION_FILE_PATH not in os.environ. It must exist "
                f"in environment variables to use be able to use the Google Ads API."
            )

        self.client = GoogleAdsClient.load_from_storage(version=self.api_version)

        return OpMsg(
            data=self,
            metadata=GoogleAdsConnectorOpMetadataModel(),
            audit=GoogleAdsConnectorOpAuditModel(),
        )



import logging
from typing import Any, Dict, Literal, Optional

from pydantic import BaseModel, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.gcp_bigquery import GCPBigQueryConnOp

logger = logging.getLogger(__name__)


class ExtractJobConfigModel(BaseModel):
    destination_format: Literal["CSV", "NEWLINE_DELIMITED_JSON", "AVRO", "PARQUET"]
    # https://cloud.google.com/bigquery/docs/reference/rest/v2/Job#JobConfigurationExtract.FIELDS.compression
    compression: Literal["NONE", "DEFLATE", "GZIP", "SNAPPY", "ZSTD"]
    field_delimiter: Optional[StrictStr] = None

    class Config:
        # NOTE WE ALLOW ADDITIONAL FIELDS HERE IN CASE WE MISSED SOMETHING
        extra = "allow"
        arbitrary_types_allowed = False


class GCPBigQueryToGCSOpConfigModel(BaseOpConfigModel):
    bigquery_conn_id: StrictStr
    project: StrictStr
    dataset_id: StrictStr
    table_id: StrictStr
    destination_uri: StrictStr
    location: StrictStr
    extract_job_config: ExtractJobConfigModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPBigQueryToGCSOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPBigQueryToGCSOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPBigQueryToGCSOp(BaseConsumerOp):
    OP_TYPE = "consumer.gcp.bigquery.extract_table"
    OP_CONFIG_MODEL = GCPBigQueryToGCSOpConfigModel
    OP_METADATA_MODEL = GCPBigQueryToGCSOpMetadataModel
    OP_AUDIT_MODEL = GCPBigQueryToGCSOpAuditModel

    templated_fields = None

    def __init__(
        self,
        bigquery_conn_id: str,
        project: str,
        dataset_id: str,
        table_id: str,
        destination_uri: str,
        location: str,
        extract_job_config: Dict[str, Any],
        **kwargs,
    ):
        """."""
        self.bigquery_conn_id = bigquery_conn_id
        self.project = project
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.destination_uri = destination_uri
        self.location = location
        self.extract_job_config = extract_job_config

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.conn: GCPBigQueryConnOp = op_manager.get_connector(
            op_id=self.bigquery_conn_id
        )

    def run(self, time_step: TimeStep, msg: OpMsg) -> OpMsg:
        """."""
        logger.info("GCPBigQueryToGCSOp: Running.")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="PangresDFToSQLDBOpOp.run:"
        )

        logger.info(f"")
        self.conn.extract_table(
            project=self.project,
            dataset_id=self.dataset_id,
            table_id=self.table_id,
            destination_uri=self.destination_uri,
            location=self.location,
            extract_job_config=self.extract_job_config,
        )

        return OpMsg(
            data=None,
            metadata=GCPBigQueryToGCSOpMetadataModel(),
            audit=GCPBigQueryToGCSOpAuditModel(),
        )



import json
import logging
from typing import Generator, Optional

import pandas as pd
from google.protobuf import json_format
from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.google_ads import GoogleAdsConnectorOp

logger = logging.getLogger(__name__)


class GoogleAdsSearchQueryGeneratorOpConfigModel(BaseOpConfigModel):
    google_ads_conn_id: StrictStr
    customer_id: StrictStr
    query: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAdsSearchQueryGeneratorOpMetadataModel(BaseOpMetadataModel):
    customer_id: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAdsSearchQueryGeneratorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAdsSearchQueryGeneratorOp(BaseGeneratorOp):
    OP_TYPE = "generator.google_ads.search_query"
    OP_CONFIG_MODEL = GoogleAdsSearchQueryGeneratorOpConfigModel
    OP_METADATA_MODEL = GoogleAdsSearchQueryGeneratorOpMetadataModel
    OP_AUDIT_MODEL = GoogleAdsSearchQueryGeneratorOpAuditModel

    templated_fields = None

    def __init__(
        self,
        google_ads_conn_id: str,
        customer_id: str,
        query: str,
        **kwargs,
    ):
        self.google_ads_conn_id = google_ads_conn_id
        self.customer_id = customer_id
        self.query = query

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.google_ads_conn: GoogleAdsConnectorOp = op_manager.get_connector(
            op_id=self.google_ads_conn_id
        )
        self.client = self.google_ads_conn.client
        self.google_ads_service = self.client.get_service("GoogleAdsService")

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"GoogleAdsSearchQueryGeneratorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="GoogleAdsSearchQueryGeneratorOp.run:",
        )

        etl_request_start_ts = pd.Timestamp.now(tz="UTC").isoformat()

        response = self.google_ads_service.search(
            customer_id=self.customer_id, query=self.query
        )

        etl_response_end_ts = pd.Timestamp.now(tz="UTC").isoformat()

        # Reference: https://stackoverflow.com/questions/55458027/convert-google-ads-api-googleadsrow-to-json
        final_records = [
            {
                "data": json.loads(json_format.MessageToJson(row._pb)),
                "etl_metadata": {
                    "time_step": time_step.to_json_dict(),
                    "etl_request_start_ts": etl_request_start_ts,
                    "etl_response_end_ts": etl_response_end_ts,
                    "query": self.query,
                    "customer_id": self.customer_id,
                },
            }
            for row in response
        ]

        logger.info(
            f"GoogleAdsSearchQueryGeneratorOp.run: Done running. {len(final_records)=}"
        )

        yield OpMsg(
            data=final_records,
            metadata=GoogleAdsSearchQueryGeneratorOpMetadataModel(
                customer_id=self.customer_id
            ),
            audit=GoogleAdsSearchQueryGeneratorOpAuditModel(),
        )



import json
import logging
import pathlib
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, StrictBool, StrictInt, StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.ops.consumer_ops.shell.shell_run import ShellRunConsumerOp

logger = logging.getLogger(__name__)


class DBTKwargsModel(BaseModel):
    project_dir: StrictStr
    profiles_dir: StrictStr
    profile: StrictStr
    target: StrictStr
    vars: Dict[StrictStr, Any] = Field(default_factory=dict)
    threads: StrictInt = 1

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DBTRunConsumerOpConfigModel(BaseOpConfigModel):
    dbt_kwargs: DBTKwargsModel
    models_include: conlist(StrictStr, min_items=1)
    models_exclude: Optional[conlist(StrictStr, min_items=1)] = None
    log_dbt_logs: StrictBool = False
    log_dbt_run_results: StrictBool = False

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DBTRunConsumerOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DBTRunConsumerOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DBTRunConsumerOp(BaseConsumerOp):
    OP_TYPE = "consumer.dbt.run"
    OP_CONFIG_MODEL = DBTRunConsumerOpConfigModel
    OP_METADATA_MODEL = DBTRunConsumerOpMetadataModel
    OP_AUDIT_MODEL = DBTRunConsumerOpAuditModel

    templated_fields = None

    def __init__(
        self,
        dbt_kwargs: Dict[str, Any],
        models_include: List[str],
        models_exclude: Optional[List[str]] = None,
        log_dbt_logs: bool = False,
        log_dbt_run_results: bool = False,
        **kwargs,
    ):
        self.dbt_kwargs = dbt_kwargs
        self.models_include = models_include
        self.models_exclude = models_exclude if models_exclude else []
        self.log_dbt_logs = log_dbt_logs
        self.log_dbt_run_results = log_dbt_run_results

        self.templated_fields = kwargs.get("templated_fields", [])

        # Will be populated after render_fields is called in run()
        self.dbt_kwargs_model: Any = None

    def run(self, time_step: TimeStep, msg: Optional[OpMsg] = None) -> OpMsg:
        """."""
        logger.info(f"DBTRunConsumerOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="DBTRunConsumerOp.run:"
        )

        self.dbt_kwargs_model = DBTKwargsModel(**self.dbt_kwargs)

        models_include_str = "--select " + " ".join(self.models_include)

        if self.models_exclude:
            models_exclude_str = "--exclude " + " ".join(self.models_exclude)
        else:
            models_exclude_str = " "

        dbt_run_shell = ShellRunConsumerOp(
            cmd=(
                f"dbt "
                f"--no-anonymous-usage-stats "
                f"run "
                f"--project-dir {self.dbt_kwargs_model.project_dir} "
                f"--profiles-dir {self.dbt_kwargs_model.profiles_dir} "
                f"--profile {self.dbt_kwargs_model.profile} "
                f"--target {self.dbt_kwargs_model.target} "
                f"--vars '{json.dumps(self.dbt_kwargs_model.vars)}' "
                f"--threads {self.dbt_kwargs_model.threads} "
                f"{models_include_str} "
                f"{models_exclude_str} "
            ),
            accepted_exit_codes=[0],
            log_cmd=True,
            log_stdout=True,
        )
        dbt_run_shell.run(time_step=time_step)

        if self.log_dbt_logs:
            logs_dir = pathlib.Path(self.dbt_kwargs_model.project_dir) / "logs"
            logs_files = list(
                sorted([x for x in logs_dir.glob("*") if x.is_file()], reverse=True)
            )
            for log_file in logs_files:
                logger.info(
                    f"DBTRunConsumerOp.run: Logging dbt log file: {str(log_file)}: \n"
                    f"{log_file.read_text()}"
                )

        if self.log_dbt_run_results:
            run_results_json = (
                pathlib.Path(self.dbt_kwargs_model.project_dir)
                / "target"
                / "run_results.json"
            )
            logger.info(
                f"DBTRunConsumerOp.run: Logging run_results_json file: "
                f"{str(run_results_json)}: \n"
                f"{run_results_json.read_text()}"
            )

        return OpMsg(
            data=None,
            metadata=DBTRunConsumerOpMetadataModel(),
            audit=DBTRunConsumerOpAuditModel(),
        )



import logging
import time

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class MockConsumerOpConfigModel(BaseOpConfigModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MockConsumerOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MockConsumerOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MockConsumerOp(BaseConsumerOp):
    OP_TYPE = "consumer.mock"
    OP_CONFIG_MODEL = MockConsumerOpConfigModel
    OP_METADATA_MODEL = MockConsumerOpMetadataModel
    OP_AUDIT_MODEL = MockConsumerOpAuditModel

    templated_fields = None

    def __init__(self, **kwargs):
        self.templated_fields = kwargs.get("templated_fields", [])

    def run(self, time_step: TimeStep, msg: OpMsg) -> OpMsg:
        """."""
        logger.info(f"MockConsumerOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="MockConsumerOp.run:"
        )

        return OpMsg(
            data=None,
            metadata=MockConsumerOpMetadataModel(),
            audit=MockConsumerOpAuditModel(),
        )



"""."""
import logging

from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class AirflowGCPHookConnOpConfigModel(BaseOpConfigModel):
    airflow_conn_id: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowGCPHookConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowGCPHookConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowGCPHookConnOp(BaseConnectorOp):
    OP_TYPE = "connector.hooks.airflow_gcp_hook"
    OP_CONFIG_MODEL = AirflowGCPHookConnOpConfigModel
    OP_METADATA_MODEL = AirflowGCPHookConnOpMetadataModel
    OP_AUDIT_MODEL = AirflowGCPHookConnOpAuditModel

    templated_fields = None

    def __init__(self, airflow_conn_id: str, **kwargs):
        self.airflow_conn_id = airflow_conn_id
        self.templated_fields = kwargs.get("templated_fields", [])

        from airflow.providers.google.common.hooks.base_google import GoogleBaseHook

        self.hook = GoogleBaseHook(gcp_conn_id=self.airflow_conn_id)

    def run(self) -> OpMsg:
        """.

        With the returned `self.analytics` object you can do:

        # ------------------------------------------------------------------------------
        # For Google Analytics
        # https://developers.google.com/analytics/devguides/reporting/core/v4/quickstart/service-py
        # ------------------------------------------------------------------------------
        import google.auth.credentials
        from googleapiclient.discovery import build
        credentials: google.auth.credentials.Credentials = (
            google_base_hook.get_credentials()
        )
        self.analytics = build("analyticsreporting", "v4", credentials=credentials)
        return analytics.reports().batchGet(
          body={
            'reportRequests': [
            {
              'viewId': VIEW_ID,
              'dateRanges': [{'startDate': '7daysAgo', 'endDate': 'today'}],
              'metrics': [{'expression': 'ga:sessions'}],
              'dimensions': [{'name': 'ga:country'}]
            }]
          }
        ).execute()
        """
        logger.info(f"AirflowGCPHookConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="AirflowGCPHookConnOp.run:"
        )

        return OpMsg(
            data=self.hook,
            metadata=AirflowGCPHookConnOpMetadataModel(),
            audit=AirflowGCPHookConnOpAuditModel(),
        )



import logging
from typing import Any, Dict, Generator, List, Optional, Tuple

import pandas as pd
from pydantic import StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp
from nr_ops.ops.ops.generator_ops.blade.get_token import BladeGetTokenOp

logger = logging.getLogger(__name__)


class BladeOrdersViewGoodsoutOpConfigModel(BaseOpConfigModel):
    http_conn_id: StrictStr
    blade_get_token_op_id: StrictStr
    goodsout_id: StrictStr
    accepted_status_codes: Optional[conlist(int, min_items=1)] = None
    remove_pii: bool = True

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeOrdersViewGoodsoutOpMetadataModel(BaseOpMetadataModel):
    goodsout_id: str
    rate_limit_requests_remaining: int

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeOrdersViewGoodsoutOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeOrdersViewGoodsoutOp(BaseGeneratorOp):
    """."""

    OP_TYPE = "generator.blade.orders.view_goodsout"
    OP_CONFIG_MODEL = BladeOrdersViewGoodsoutOpConfigModel
    OP_METADATA_MODEL = BladeOrdersViewGoodsoutOpMetadataModel
    OP_AUDIT_MODEL = BladeOrdersViewGoodsoutOpAuditModel

    templated_fields = None

    def __init__(
        self,
        http_conn_id: str,
        blade_get_token_op_id: str,
        goodsout_id: str,
        accepted_status_codes: Optional[List[int]] = None,
        remove_pii: bool = True,
        **kwargs,
    ):
        """."""
        self.http_conn_id = http_conn_id
        self.blade_get_token_op_id = blade_get_token_op_id
        self.goodsout_id = goodsout_id
        self.accepted_status_codes = accepted_status_codes
        self.remove_pii = remove_pii

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(op_id=self.http_conn_id)

        self.blade_get_token_op: BladeGetTokenOp = op_manager.op.get_op(
            op_id=self.blade_get_token_op_id
        ).op_obj

    def get_goodsout(self, goodsout_id: str) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """."""
        # Example URL:
        # https://api.blade.co.uk/v1/orders/goodsouts/123456789
        # ?expand=*
        url = f"{self.http_conn.base_url}/v1/orders/goodsouts/{goodsout_id}?expand=*"
        logger.info(
            f"BladeOrdersViewGoodsoutOp.get_page: Fetching {goodsout_id=} with {url=}"
        )

        etl_request_start_ts = pd.Timestamp.now(tz="UTC").isoformat()
        status_code, output_json = self.http_conn.call(
            method="get",
            url=url,
            requests_kwargs={
                "headers": {
                    "Content-Type": "application/json",
                    # Token gets refreshed if it has expired.
                    "Access-Token": self.blade_get_token_op.token,
                }
            },
            accepted_status_codes=self.accepted_status_codes,
            return_type="json",
        )
        etl_response_end_ts = pd.Timestamp.now(tz="UTC").isoformat()

        etl_metadata_json = {
            "etl_request_start_ts": etl_request_start_ts,
            "etl_response_end_ts": etl_response_end_ts,
        }

        return etl_metadata_json, output_json

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"BladeOrdersViewGoodsoutOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="BladeOrdersViewGoodsoutOp.run:"
        )

        logger.info(f"BladeOrdersViewGoodsoutOp.run: Fetching {self.goodsout_id=}.")

        etl_metadata_json, output_json = self.get_goodsout(goodsout_id=self.goodsout_id)
        rate_limit_requests_remaining = output_json["meta"]["request_rate_limit"][
            "requests_remaining"
        ]

        logger.info(
            f"BladeOrdersViewGoodsoutOp.run: Done fetching {self.goodsout_id=}. "
            f"{rate_limit_requests_remaining=}"
        )

        if self.remove_pii:
            logger.info(f"BladeOrdersViewGoodsoutOp.run: Redacting PII from the data.")
            # Redact PII from the data in place.
            for key in [
                "first_name",
                "last_name",
                "email",
                "mobile",
                "telephone",
                "secondary_telephone",
                "address_one",
                "address_two",
                "address_three",
                "company",
                # Specific to view_goodsout
                "vat_number",
            ]:
                output_json["data"]["shipping_address"][
                    key
                ] = "REDACTED_BY_ETL_BEFORE_STORAGE"

            for key in ["delivery_instructions", "gift_message"]:
                output_json["data"][key] = "REDACTED_BY_ETL_BEFORE_STORAGE"

            logger.info(
                f"BladeOrdersViewGoodsoutOp.run: Done redacting PII from the data."
            )

            logger.info(
                f"BladeOrdersViewGoodsoutOp.run: Yielding results for page "
                f"{self.goodsout_id=}."
            )

        ################################################################################
        # NOTE: Adding `etl_metadata` into response json (output_json).
        # * This is to add some metadata about the ETL process, primarily some
        #   timestamps which may help with creating event stores/logs.
        # * This may also help with incremental modelling.
        # * DO NOT USE this information when calculating dedup_uuid.
        ################################################################################
        etl_metadata_json["time_step"] = time_step.to_json_dict()
        etl_metadata_json["blade"] = {"meta": output_json.get("meta", {})}
        output_json["etl_metadata"] = etl_metadata_json

        yield OpMsg(
            data=output_json,
            metadata=BladeOrdersViewGoodsoutOpMetadataModel(
                goodsout_id=self.goodsout_id,
                rate_limit_requests_remaining=rate_limit_requests_remaining,
            ),
            audit=BladeOrdersViewGoodsoutOpAuditModel(),
        )



import base64
import bz2
import collections
import copy
import datetime
import gzip
import hashlib
import importlib
import io
import itertools
import json
import logging
import math
import pathlib
import pickle
import pprint
import random
import re
import shutil
import tarfile
import time
import uuid
import zipfile
import zlib
from functools import partial

import numpy as np
import pandas as pd
import requests
import sklearn
import sklearn.datasets
import sklearn.linear_model
import sklearn.metrics
import sklearn.model_selection
import sklearn.pipeline
import sklearn.preprocessing
import sklearn.svm
import zstandard as zstd

from nr_ops.utils.eval.custom_functions import (
    custom_generate_batches_from_list,
    custom_raise_exception,
)

logger = logging.getLogger(__name__)


EVAL_GLOBALS = {
    "base64": base64,
    "bz2": bz2,
    "collections": collections,
    "copy": copy,
    "datetime": datetime,
    "gzip": gzip,
    "hashlib": hashlib,
    "importlib": importlib,
    "io": io,
    "itertools": itertools,
    "json": json,
    "logging": logging,
    "math": math,
    "pathlib": pathlib,
    "pickle": pickle,
    "pprint": pprint,
    "random": random,
    "re": re,
    "shutil": shutil,
    "tarfile": tarfile,
    "time": time,
    "uuid": uuid,
    "zipfile": zipfile,
    "zlib": zlib,
    "zstd": zstd,
    "partial": partial,
    # third-party
    "np": np,
    "pd": pd,
    "requests": requests,
    # third-party: sklearn
    "sklearn": sklearn,
    "sklearn.datasets": sklearn.datasets,
    "sklearn.linear_model": sklearn.linear_model,
    "sklearn.metrics": sklearn.metrics,
    "sklearn.model_selection": sklearn.model_selection,
    "sklearn.pipeline": sklearn.pipeline,
    "sklearn.preprocessing": sklearn.preprocessing,
    "sklearn.svm": sklearn.svm,
    # custom_functions
    "custom_raise_exception": custom_raise_exception,
    "custom_generate_batches_from_list": custom_generate_batches_from_list,
}



import argparse
import logging
import traceback
from logging.config import dictConfig

import ruamel.yaml

from nr_ops.config.main_config import MainConfigModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.op import Op
from nr_ops.ops.op_manager import init_global_op_manager
from nr_ops.utils.logging.dict_config import LoggingDictConfigModel

logger = logging.getLogger(__name__)


def core(
    config,
    root_msg: OpMsg,
    return_generator: bool = False,
):
    """Run core logic after logging has been initialized."""

    # Validate config
    logger.info("Validating config against MainConfigModel...")
    config_model = MainConfigModel(**config)
    config_model.root_op.validate_unique_ids()
    logger.info("Done validating config against MainConfigModel...")

    # Configure user provided logging.
    logger.info("Configuring user provided logging config...")
    while logger.handlers:
        logger.handlers.pop()

    dictConfig(config=config_model.logging.dict())
    logger.info("Configured user provided logging!")

    # Start with nr_ops...

    # Global op manager needs to be initialized before any ops are created.
    op_manager = init_global_op_manager()
    op_manager.set_env_vars(env_vars=config_model.set_env_vars)
    op_manager.import_env_vars(env_vars=config_model.import_env_vars)

    # ----------------------------------------------------------------------------------
    # Initialize Connectors (if any)
    # ----------------------------------------------------------------------------------
    if config_model.connector_op is not None:
        connector = Op(**config_model.connector_op.dict())
        for _ in connector.run():
            pass

    schedule_op = Op(**config_model.schedule_op.dict())

    root_op = Op(**config["root_op"])

    logger.info(f"main: Running schedule_op")
    for time_step_index, time_step_msg in enumerate(schedule_op.run()):
        time_step = time_step_msg.data
        time_step.index = time_step_index

        logger.info(f"{'#' * 80}")
        logger.info(f"main: Running root_op for {time_step=}")
        logger.info(f"{'#' * 80}")

        if return_generator:
            # This can be used by a server to collect the results (if there's any
            # results)
            return root_op.run(
                time_step=time_step,
                msg=root_msg,
            )
        else:
            for _msg in root_op.run(
                time_step=time_step,
                msg=root_msg,
            ):
                # This exhausts the generator of `root_op` which is the core behavior
                # of nr-ops CLI.
                pass

    logger.info("Done running nr_ops.main.core!")


def run():
    """Main entrypoint for the nr_ops package."""
    msg = (
        "Running nr_ops.main.run(), no logging initialized yet. Attempting to "
        "configure default logging..."
    )
    print(msg)
    logger.info(msg)

    # Configure default logging until the user provided logging config is loaded.
    dictConfig(config=LoggingDictConfigModel().dict())
    logger.info("Configured default logging...")

    try:
        # Parse command line arguments.
        parser = argparse.ArgumentParser()
        parser.add_argument("--config", help="Path to the config file.")
        args = parser.parse_args()
        logger.info("Parsed CLI arguments...")

        # Read config
        yaml = ruamel.yaml.YAML(typ="safe", pure=True)

        with open(args.config, "r") as f:
            config = yaml.load(f)

        logger.info("Parsed config...")
        logger.info("Running core...")

        root_msg = OpMsg(data=None, metadata={}, audit={})
        core(config=config, root_msg=root_msg)
        logger.info("Done running core...")
    except Exception:
        logger.exception(
            f"Exception occurred in nr_ops.main.core()!\n{traceback.format_exc()}"
        )
        raise

    logger.info("Done running nr_ops.main.run!")



import logging
from typing import Generator, Literal, Optional

from pydantic import StrictStr, validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.gcp_gcs import GCPGCSConnOp

logger = logging.getLogger(__name__)


class GCPGCSGetKeyOpConfigModel(BaseOpConfigModel):
    gcs_conn_id: StrictStr
    bucket: StrictStr
    key: StrictStr
    output_type: Literal["bytes", "bytesio"]

    @validator("key", pre=False)
    def validate_key(cls, key: str):
        if key.endswith("/"):
            raise ValueError(f"{key=} must not end with a '/'")

        if key.startswith("/"):
            raise ValueError(f"{key=} must not start with a '/'")

        return key

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSGetKeyOpMetadataModel(BaseOpMetadataModel):
    bucket: StrictStr
    key: StrictStr
    output_type: Literal["bytes", "bytesio"]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSGetKeyOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSGetKeyOp(BaseGeneratorOp):
    OP_TYPE = "generator.gcp.gcs.get_key"
    OP_CONFIG_MODEL = GCPGCSGetKeyOpConfigModel
    OP_METADATA_MODEL = GCPGCSGetKeyOpMetadataModel
    OP_AUDIT_MODEL = GCPGCSGetKeyOpAuditModel

    templated_fields = None

    def __init__(
        self,
        gcs_conn_id: str,
        bucket: str,
        key: str,
        output_type: Literal["bytes", "bytesio"],
        **kwargs,
    ):
        """."""
        self.gcs_conn_id = gcs_conn_id
        self.bucket = bucket
        self.key = key
        self.output_type = output_type
        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.gcs_conn: GCPGCSConnOp = op_manager.get_connector(op_id=self.gcs_conn_id)

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"GCPGCSGetKeyOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="GCPGCSGetKeyOp.run:"
        )

        logger.info(
            f"GCPGCSGetKeyOp.run: Getting key {self.key=} with {self.output_type=}"
        )

        output = self.gcs_conn.get_key(
            bucket=self.bucket, key=self.key, output_type=self.output_type
        )

        if self.output_type == "bytes":
            logger.info(
                f"GCPGCSGetKeyOp.run: Download with {len(output):,} bytes "
                f"i.e. approx {len(output)//1000//1000:,} MiB"
            )

        logger.info(
            f"GCPGCSGetKeyOp.run: Return data with {self.output_type=} for {self.key=}"
        )
        yield OpMsg(
            data=output,
            metadata=GCPGCSGetKeyOpMetadataModel(
                bucket=self.bucket,
                key=self.key,
                output_type=self.output_type,
            ),
            audit=GCPGCSGetKeyOpAuditModel(),
        )



import logging
from io import StringIO
from typing import Any, Dict, Generator, Literal, Optional

import pandas as pd

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class PandasReadGenericOpConfigModel(BaseOpConfigModel):
    input_as_argument: bool = True
    read_type: Literal[
        "read_csv",
        "read_excel",
        "read_json",
        "read_parquet",
        "read_sql",
    ]
    read_kwargs: Dict[str, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PandasReadGenericOpMetadataModel(BaseOpMetadataModel):
    read_kwargs: Dict[str, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PandasReadGenericOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PandasReadGenericOp(BaseGeneratorOp):
    OP_TYPE = "generator.pandas.read_generic"
    OP_CONFIG_MODEL = PandasReadGenericOpConfigModel
    OP_METADATA_MODEL = PandasReadGenericOpMetadataModel
    OP_AUDIT_MODEL = PandasReadGenericOpAuditModel

    templated_fields = None

    def __init__(
        self,
        read_type: str,
        read_kwargs: Dict[str, Any],
        input_as_argument: bool = True,
        **kwargs,
    ):
        self.read_type = read_type
        self.read_kwargs = read_kwargs
        self.input_as_argument = input_as_argument

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"PandasReadGenericOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="PandasReadGenericOp.run:"
        )

        read_func = getattr(pd, self.read_type)
        if self.input_as_argument:
            output = read_func(msg.data, **self.read_kwargs)
        else:
            output = read_func(**self.read_kwargs)

        # Special logging for pandas DataFrames
        buf = StringIO()
        output.info(buf=buf)
        buf.seek(0)
        logger.info(f"PandasReadGenericOp.run: DataFrame info:\n{buf.read()}")

        yield OpMsg(
            data=output,
            metadata=PandasReadGenericOpMetadataModel(read_kwargs=self.read_kwargs),
            audit=PandasReadGenericOpAuditModel(),
        )



import logging
from typing import Generator, Optional

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class OngoingGenericRemovePiiOpConfigModel(BaseOpConfigModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OngoingGenericRemovePiiOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OngoingGenericRemovePiiOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OngoingGenericRemovePiiOp(BaseGeneratorOp):
    """."""

    OP_TYPE = "generator.ongoing.generic.remove_pii"
    OP_CONFIG_MODEL = OngoingGenericRemovePiiOpConfigModel
    OP_METADATA_MODEL = OngoingGenericRemovePiiOpMetadataModel
    OP_AUDIT_MODEL = OngoingGenericRemovePiiOpAuditModel

    templated_fields = None

    def __init__(
        self,
        **kwargs,
    ):
        """."""
        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"OngoingGenericRemovePiiOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="OngoingGenericRemovePiiOp.run:",
        )

        input_records = msg.data

        if not isinstance(input_records, list):
            raise NotImplementedError("Input records must be a list of dicts.")

        final_records = []

        logger.info(f"OngoingGenericRemovePiiOp.run: Redacting PII from the data.")

        keys_to_redact = [
            "Name",
            "Address",
            "Address2",
            "Address3",
            "TelePhone",
            "MobilePhone",
            "Email",
            "Remark",
        ]
        for record in input_records:
            for key in keys_to_redact:
                if "Consignee" in record["data"]:
                    record["data"]["Consignee"][key] = "REDACTED_BY_ETL_BEFORE_STORAGE"

            if "OrderInfo" in record["data"]:
                if "EmailNotification" in record["data"]["OrderInfo"]:
                    record["data"]["OrderInfo"]["EmailNotification"][
                        "Value"
                    ] = "REDACTED_BY_ETL_BEFORE_STORAGE"

                if "TelephoneNotification" in record["data"]["OrderInfo"]:
                    record["data"]["OrderInfo"]["TelephoneNotification"][
                        "Value"
                    ] = "REDACTED_BY_ETL_BEFORE_STORAGE"

                if "SMSNotification" in record["data"]["OrderInfo"]:
                    record["data"]["OrderInfo"]["SMSNotification"][
                        "Value"
                    ] = "REDACTED_BY_ETL_BEFORE_STORAGE"

            # record['data'] has been updated in place.
            final_records.append(record)

        logger.info(f"OngoingGenericRemovePiiOp.run: Done redacting PII from the data.")

        yield OpMsg(
            data=final_records,
            metadata=OngoingGenericRemovePiiOpMetadataModel(),
            audit=OngoingGenericRemovePiiOpAuditModel(),
        )



import logging
from typing import List

from pydantic import StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class GCPServiceAccountFromFileConnOpConfigModel(BaseOpConfigModel):
    key_file_path: StrictStr
    scopes: conlist(StrictStr, min_items=0)

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPServiceAccountFromFileConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPServiceAccountFromFileConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPServiceAccountFromFileConnOp(BaseConnectorOp):
    OP_TYPE = "connector.hooks.gcp_service_account_from_file"
    OP_CONFIG_MODEL = GCPServiceAccountFromFileConnOpConfigModel
    OP_METADATA_MODEL = GCPServiceAccountFromFileConnOpMetadataModel
    OP_AUDIT_MODEL = GCPServiceAccountFromFileConnOpAuditModel

    templated_fields = None

    def __init__(self, key_file_path: str, scopes: List[str], **kwargs):
        self.key_file_path = key_file_path
        self.scopes = scopes
        self.templated_fields = kwargs.get("templated_fields", [])

        # Will be populated by run()
        self.credentials = None

    def run(self) -> OpMsg:
        """.

        # ------------------------------------------------------------------------------
        # EXAMPLES - Google Analytics Reporting API v4
        # https://googleapis.github.io/google-api-python-client/docs/dyn/analyticsreporting_v4.html
        # ------------------------------------------------------------------------------
        self.api = build("analyticsreporting", "v4", credentials=credentials)
        return self.api.reports().batchGet(
          body={
            'reportRequests': [
            {
              'viewId': VIEW_ID,
              'dateRanges': [{'startDate': '7daysAgo', 'endDate': 'today'}],
              'metrics': [{'expression': 'ga:sessions'}],
              'dimensions': [{'name': 'ga:country'}]
            }]
          }
        ).execute()

        """
        logger.info(f"GCPServiceAccountFromFileConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="GCPServiceAccountFromFileConnOp.run:"
        )

        # Source: https://googleapis.github.io/google-api-python-client/docs/dyn/analyticsreporting_v4.html
        from oauth2client.service_account import ServiceAccountCredentials

        self.credentials = ServiceAccountCredentials.from_json_keyfile_name(
            self.key_file_path, self.scopes
        )

        # https://developers.google.com/analytics/devguides/reporting/core/v4/quickstart/service-py
        return OpMsg(
            data=self,
            metadata=GCPServiceAccountFromFileConnOpMetadataModel(),
            audit=GCPServiceAccountFromFileConnOpAuditModel(),
        )



import logging
import time
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple

import pandas as pd
from pydantic import StrictBool, StrictStr, conlist, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp

logger = logging.getLogger(__name__)


class AirflowDagRunGetDagRunOpConfigModel(BaseOpConfigModel):
    http_conn_id: StrictStr
    dag_id: StrictStr
    dag_run_id: StrictStr
    accepted_status_codes: Optional[conlist(int, min_items=1)] = None
    include_etl_metadata: StrictBool = False
    requests_kwargs: Optional[Dict[str, Any]] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowDagRunGetDagRunOpMetadataModel(BaseOpMetadataModel):
    etl_request_start_ts: StrictStr
    etl_response_end_ts: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowDagRunGetDagRunOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowDagRunGetDagRunOp(BaseGeneratorOp):
    """."""

    OP_TYPE = "generator.airflow.dagruns.get_dagrun"
    OP_CONFIG_MODEL = AirflowDagRunGetDagRunOpConfigModel
    OP_METADATA_MODEL = AirflowDagRunGetDagRunOpMetadataModel
    OP_AUDIT_MODEL = AirflowDagRunGetDagRunOpAuditModel

    templated_fields = None

    def __init__(
        self,
        http_conn_id: str,
        dag_id: str,
        dag_run_id: str,
        accepted_status_codes: Optional[List[int]] = None,
        include_etl_metadata: bool = False,
        requests_kwargs: Optional[Dict[str, any]] = None,
        **kwargs,
    ):
        """."""
        self.http_conn_id = http_conn_id
        self.dag_id = dag_id
        self.dag_run_id = dag_run_id
        self.accepted_status_codes = accepted_status_codes
        self.include_etl_metadata = include_etl_metadata
        self.requests_kwargs = requests_kwargs if requests_kwargs is not None else {}

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(op_id=self.http_conn_id)

    def get_dagrun(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """."""
        url = (
            f"{self.http_conn.base_url}/api/v1/"
            f"dags/{self.dag_id}/"
            f"dagRuns/{self.dag_run_id}"
        )
        logger.info(
            f"AirflowDagRunGetDagRunOp.get_dagrun: Getting DAGRun with "
            f"{self.dag_id=} and {self.dag_run_id=} with {url=}"
        )

        etl_request_start_ts = pd.Timestamp.now(tz="UTC").isoformat()

        status_code, response_json = self.http_conn.call(
            method="get",
            url=url,
            requests_kwargs={
                "headers": {
                    "Content-Type": "application/json",
                },
                **self.requests_kwargs,
            },
            accepted_status_codes=self.accepted_status_codes,
            return_type="json",
        )
        etl_response_end_ts = pd.Timestamp.now(tz="UTC").isoformat()

        etl_metadata_json = {
            "etl_request_start_ts": etl_request_start_ts,
            "etl_response_end_ts": etl_response_end_ts,
        }

        output_json = {
            "status_code": status_code,
            "response": response_json,
        }

        return etl_metadata_json, output_json

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"AirflowDagRunGetDagRunOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="AirflowDagRunGetDagRunOp.run:"
        )

        logger.info(f"AirflowDagRunGetDagRunOp.run: Yielding results for page")

        etl_metadata_json, output_json = self.get_dagrun()

        if self.include_etl_metadata:
            output_json["etl_metadata"] = etl_metadata_json

        yield OpMsg(
            data=output_json,
            metadata=AirflowDagRunGetDagRunOpMetadataModel(
                **etl_metadata_json,
            ),
            audit=AirflowDagRunGetDagRunOpAuditModel(),
        )



import importlib.metadata
from typing import Any, Dict, List, Optional

import pkg_resources
from pydantic import BaseModel, Field, StrictStr, validator

from nr_ops.ops.op import OpModel
from nr_ops.utils.logging.dict_config import LoggingDictConfigModel


class MainConfigModel(BaseModel):
    docs: Optional[StrictStr] = None
    app_version: StrictStr
    set_env_vars: Optional[Dict[StrictStr, StrictStr]] = None
    import_env_vars: Optional[List[StrictStr]] = None
    logging: LoggingDictConfigModel = Field(default_factory=LoggingDictConfigModel)
    connector_op: Optional[OpModel] = None
    schedule_op: OpModel
    root_op: OpModel

    @validator("app_version")
    def validate_app_version(cls, app_version):
        """Validate that `app_version` in config file is the nr_ops_version."""
        nr_ops_version = pkg_resources.get_distribution("nr_ops").version
        if app_version != nr_ops_version:
            raise ValueError(
                f"{app_version=} provided in the config file does not match "
                f"{nr_ops_version=}! Please update the config file or use a different "
                f"version of nr-ops."
            )
        return app_version

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False



import logging
from io import StringIO

import pandas as pd

logger = logging.getLogger(__name__)


def log_df_info(df: pd.DataFrame, log_prefix: str = ""):
    buf = StringIO()
    df.info(buf=buf)
    buf.seek(0)
    logger.info(f"{log_prefix}DataFrame info:\n{buf.read()}")



import logging
from typing import Any, Dict, List, Literal, Optional

from googleapiclient.discovery import build
from pydantic import StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel
from nr_ops.ops.ops.connector_ops.interfaces.base import validate_hook_type_and_config

logger = logging.getLogger(__name__)


class GoogleSheetsConnOpConfigModel(BaseOpConfigModel):
    hook_type: Literal[
        "connector.hooks.gcp_service_account_from_file",
    ]
    hook_config: Dict[StrictStr, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    _validate_hook_type_and_config = root_validator(allow_reuse=True)(
        validate_hook_type_and_config
    )


class GoogleSheetsConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleSheetsConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleSheetsConnOp(BaseConnectorOp):
    OP_TYPE = "connector.google_sheets"
    OP_CONFIG_MODEL = GoogleSheetsConnOpConfigModel
    OP_METADATA_MODEL = GoogleSheetsConnOpMetadataModel
    OP_AUDIT_MODEL = GoogleSheetsConnOpAuditModel

    templated_fields = None

    def __init__(self, hook_type: str, hook_config: Dict[str, Any], **kwargs):
        self.hook_type = hook_type
        self.hook_config = hook_config
        self.templated_fields = kwargs.get("templated_fields", [])

        from nr_ops.ops.op_collection import OP_COLLECTION

        self.hook_op = OP_COLLECTION[self.hook_type](**self.hook_config)
        self.hook = self.hook_op.run().data

        # Will be populated in run()
        self.sheet_service: Any = None

    def run(self) -> OpMsg:
        """."""
        logger.info(f"GoogleSheetsConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="GoogleSheetsConnOp.run:"
        )

        # Source: https://googleapis.github.io/google-api-python-client/docs/dyn/analyticsreporting_v4.html
        if self.hook_type == "connector.hooks.gcp_service_account_from_file":
            # Type annotate for pycharm/mypy/autocomplete

            from oauth2client.service_account import ServiceAccountCredentials

            from nr_ops.ops.ops.connector_ops.hooks.gcp_service_account_from_file import (
                GCPServiceAccountFromFileConnOp,
            )

            self.hook: GCPServiceAccountFromFileConnOp
            credentials: ServiceAccountCredentials = self.hook.credentials

            service = build("sheets", "v4", credentials=credentials)

            # Init the sheet API
            self.sheet_service = service.spreadsheets()

        else:
            raise NotImplementedError()

        return OpMsg(
            data=self,
            metadata=GoogleSheetsConnOpMetadataModel(),
            audit=GoogleSheetsConnOpAuditModel(),
        )



import io
import logging
from typing import Any, Dict, List, Literal

from pydantic import StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel
from nr_ops.ops.ops.connector_ops.interfaces.base import validate_hook_type_and_config

logger = logging.getLogger(__name__)


class GCPGCSConnOpConfigModel(BaseOpConfigModel):
    hook_type: Literal[
        "connector.hooks.airflow_gcp_gcs_hook",
    ]
    hook_config: Dict[StrictStr, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    _validate_hook_type_and_config = root_validator(allow_reuse=True)(
        validate_hook_type_and_config
    )


class GCPGCSConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSConnOp(BaseConnectorOp):
    OP_TYPE = "connector.gcp.gcs"
    OP_CONFIG_MODEL = GCPGCSConnOpConfigModel
    OP_METADATA_MODEL = GCPGCSConnOpMetadataModel
    OP_AUDIT_MODEL = GCPGCSConnOpAuditModel

    templated_fields = None

    def __init__(self, hook_type: str, hook_config: Dict[str, Any], **kwargs):
        self.hook_type = hook_type
        self.hook_config = hook_config
        self.templated_fields = kwargs.get("templated_fields", [])

        from nr_ops.ops.op_collection import OP_COLLECTION

        self.hook_op = OP_COLLECTION[self.hook_type](**self.hook_config)
        self.hook = self.hook_op.run().data

        from airflow.providers.google.cloud.hooks.gcs import GCSHook

        # Will be populated in run()
        self.gcs_hook: GCSHook = None

    def list_keys(self, bucket: str, prefix: str, delimiter: str = "/") -> List[str]:
        """."""
        keys = self.gcs_hook.list(
            bucket_name=bucket,
            prefix=prefix,
            delimiter=delimiter,
        )
        return keys

    def get_key(self, bucket: str, key: str, output_type: str):
        """."""
        bytes_obj = self.gcs_hook.download(bucket_name=bucket, object_name=key)
        if output_type == "bytes":
            return bytes_obj
        elif output_type == "bytesio":
            bytesio_obj = io.BytesIO(bytes_obj)
            bytesio_obj.seek(0)
            return bytesio_obj
        else:
            raise NotImplementedError()

    def put_key(
        self,
        bucket: str,
        key: str,
        input_type: str,
        replace: bool,
        data: Any,
    ):
        if replace is True:
            raise NotImplementedError(f"{replace=} is not implemented yet.")

        if input_type == "bytes":
            self.gcs_hook.upload(
                bucket_name=bucket,
                object_name=key,
                data=data,
                # TODO: Make this configurable after figuring out what effect
                #  it has.
                encoding="utf-8",
            )
        elif input_type == "bytesio":
            self.gcs_hook.upload(
                bucket_name=bucket,
                object_name=key,
                # NOTE: There is no file_obj based upload method in GCSHook.
                # Therefore, we have to read the data as bytes.
                data=data.read(),
                # TODO: Make this configurable after figuring out what effect it has
                encoding="utf-8",
            )
        else:
            raise NotImplementedError()

    def delete_key(
        self,
        bucket: str,
        key: str,
    ):
        self.gcs_hook.delete(
            bucket_name=bucket,
            object_name=key,
        )
        self.gcs_hook

    def is_key_exists(self, bucket: str, key: str) -> bool:
        """."""
        is_exists = self.gcs_hook.exists(bucket_name=bucket, object_name=key)
        return is_exists

    def get_blob(self, bucket: str, key: str) -> bool:
        """."""
        client = self.gcs_hook.get_conn()
        bucket = client.bucket(bucket)
        blob = bucket.blob(blob_name=key)
        return blob

    def run(self) -> OpMsg:
        """."""
        logger.info(f"S3ConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=None, msg=None, log_prefix="S3ConnOp.run:")

        if self.hook_type == "connector.hooks.airflow_gcp_gcs_hook":
            from airflow.providers.google.cloud.hooks.gcs import GCSHook

            self.gcs_hook: GCSHook = self.hook

        else:
            raise NotImplementedError()

        return OpMsg(
            data=self,
            metadata=GCPGCSConnOpMetadataModel(),
            audit=GCPGCSConnOpAuditModel(),
        )



import logging
import time
from typing import Any, Dict, Generator, List, Optional, Tuple

import pandas as pd
from pydantic import StrictBool, StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp

logger = logging.getLogger(__name__)


class BigComCatalogGetAllBrandsOpConfigModel(BaseOpConfigModel):
    http_conn_id: StrictStr
    store_hash: StrictStr
    accepted_status_codes: Optional[conlist(int, min_items=1)] = None
    sleep_time_between_pages: int = 5
    timeout_seconds_per_request: float = 60
    iterate_over_pages: StrictBool

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BigComCatalogGetAllBrandsOpMetadataModel(BaseOpMetadataModel):
    current_page: int
    total_pages: int
    total_records: int

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BigComCatalogGetAllBrandsOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BigComCatalogGetAllBrandsOp(BaseGeneratorOp):
    """.

    NOTES:
    - The `page` field in the metadata is 1-indexed.
    """

    OP_TYPE = "generator.bigcom.catalog.get_all_brands"
    OP_CONFIG_MODEL = BigComCatalogGetAllBrandsOpConfigModel
    OP_METADATA_MODEL = BigComCatalogGetAllBrandsOpMetadataModel
    OP_AUDIT_MODEL = BigComCatalogGetAllBrandsOpAuditModel

    templated_fields = None

    def __init__(
        self,
        http_conn_id: str,
        store_hash: str,
        iterate_over_pages: bool,
        accepted_status_codes: Optional[List[int]] = None,
        sleep_time_between_pages: int = 5,
        timeout_seconds_per_request: float = 60,
        **kwargs,
    ):
        """."""
        self.http_conn_id = http_conn_id
        self.store_hash = store_hash
        self.iterate_over_pages = iterate_over_pages
        self.sleep_time_between_pages = sleep_time_between_pages
        self.accepted_status_codes = (
            accepted_status_codes if accepted_status_codes else [200]
        )
        self.timeout_seconds_per_request = timeout_seconds_per_request

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(op_id=self.http_conn_id)

    def get_page(self, page: int) -> Tuple[int, Dict[str, Any], Dict[str, Any]]:
        """."""
        params = {}

        params["page"] = page

        # DOCS: https://developer.bigcommerce.com/docs/rest-catalog/brands#get-all-brands
        url = f"{self.http_conn.base_url}/stores/{self.store_hash}/v3/catalog/brands"
        logger.info(
            f"BigComCatalogGetAllBrandsOp.get_page: Fetching {page=} with {url=}"
        )

        etl_request_start_ts = pd.Timestamp.now(tz="UTC").isoformat()
        status_code, output_json = self.http_conn.call(
            method="get",
            url=url,
            requests_kwargs={
                "params": params,
                "headers": {
                    "Content-Type": "application/json",
                    # AUTHENTICATION HEADERS WILL BE INJECTED BY THE HTTP CONN FROM EXTRAS
                },
                "timeout": self.timeout_seconds_per_request,
            },
            accepted_status_codes=self.accepted_status_codes,
            # If status_code = 200, then the return_type is JSON
            # If status_code = 204, then the return_type is TEXT
            # Therefore we use text as the default return_type. We will convert to JSON if needed.
            return_type="json",
        )
        etl_response_end_ts = pd.Timestamp.now(tz="UTC").isoformat()

        etl_metadata_json = {
            "etl_request_start_ts": etl_request_start_ts,
            "etl_response_end_ts": etl_response_end_ts,
        }

        return status_code, output_json, etl_metadata_json

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"BigComCatalogGetAllBrandsOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="BigComCatalogGetAllBrandsOp.run:",
        )

        final_records = []
        page, total_records, total_pages = 1, 0, None
        while True:
            logger.info(f"BigComCatalogGetAllBrandsOp.run: Fetching {page=}.")

            status_code, output_json, etl_metadata_json = self.get_page(page=page)
            total_pages = output_json["meta"]["pagination"]["total_pages"]
            n_records = len(output_json["data"])
            total_records += n_records

            logger.info(
                f"BigComCatalogGetAllBrandsOp.run: Done fetching {page=}. "
                f"Fetched {n_records=} this request. "
                f"Fetched {total_records=} records so far. "
            )

            ############################################################################
            # NOTE: Adding `etl_metadata` into response json (output_json).
            # * This is to add some metadata about the ETL process, primarily some
            #   timestamps which may help with creating event stores/logs.
            # * This may also help with incremental modelling.
            # * DO NOT USE this information when calculating dedup_uuid.
            ############################################################################
            etl_metadata_json["meta"] = output_json["meta"]
            etl_metadata_json["time_step"] = time_step.to_json_dict()
            records = [
                {
                    "data": record,
                    "etl_metadata": etl_metadata_json,
                }
                for record in output_json["data"]
            ]
            final_records.extend(records)

            # Yield current page's results
            # If true, yield current records. Afterwards set final_records to empty list
            # If true, make sure to not yield records outside of loop.
            if self.iterate_over_pages:
                yield OpMsg(
                    data=final_records,
                    metadata=BigComCatalogGetAllBrandsOpMetadataModel(
                        current_page=page,
                        total_pages=total_pages,
                        total_records=total_records,
                    ),
                    audit=BigComCatalogGetAllBrandsOpAuditModel(),
                )
                final_records = []

            # Break look if page >= total_pages
            # Don't sleep on the last page
            if page >= total_pages:
                logger.info(
                    f"BigComCatalogGetAllBrandsOp.run: Fetched {total_pages=}. "
                    f"Yielding records."
                )
                break

            logger.info(
                f"BigComCatalogGetAllBrandsOp.run: Sleeping for "
                f"{self.sleep_time_between_pages} seconds between pages. "
                f"Next page is {page+1}."
            )
            time.sleep(self.sleep_time_between_pages)

            # IMPORTANT: The page number is 1-indexed.
            page += 1

        # Only yield records if iterate_over_pages is false.
        if self.iterate_over_pages:
            logger.info(
                f"BigComCatalogGetAllBrandsOp.run: {self.iterate_over_pages=}. "
                f"Finished yielding all records, once per page."
            )

        else:
            logger.info(
                f"BigComCatalogGetAllBrandsOp.run: {self.iterate_over_pages=}. "
                f"Yielding all pages at once now. {len(final_records)=}"
            )
            yield OpMsg(
                data=final_records,
                metadata=BigComCatalogGetAllBrandsOpMetadataModel(
                    current_page=page,
                    total_pages=total_pages,
                    total_records=total_records,
                ),
                audit=BigComCatalogGetAllBrandsOpAuditModel(),
            )



import logging
import time
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple

import pandas as pd
from pydantic import StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp
from nr_ops.ops.ops.generator_ops.blade.get_token import BladeGetTokenOp

logger = logging.getLogger(__name__)


class BladeProductsListVariationsOpConfigModel(BaseOpConfigModel):
    http_conn_id: StrictStr
    blade_get_token_op_id: StrictStr
    accepted_status_codes: Optional[conlist(int, min_items=1)] = None
    sleep_time_between_pages: int = 5
    sort_by: Optional[
        Literal[
            "created",
        ]
    ] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeProductsListVariationsOpMetadataModel(BaseOpMetadataModel):
    page: int
    total_pages: int
    n_records: int
    n_records_processed: int
    total_records: int
    rate_limit_requests_remaining: int

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeProductsListVariationsOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BladeProductsListVariationsOp(BaseGeneratorOp):
    """.

    NOTES:
    - The `page` field in the metadata is 1-indexed.
    """

    OP_TYPE = "generator.blade.products.list_variations"
    OP_CONFIG_MODEL = BladeProductsListVariationsOpConfigModel
    OP_METADATA_MODEL = BladeProductsListVariationsOpMetadataModel
    OP_AUDIT_MODEL = BladeProductsListVariationsOpAuditModel

    templated_fields = None

    def __init__(
        self,
        http_conn_id: str,
        blade_get_token_op_id: str,
        accepted_status_codes: Optional[List[int]] = None,
        sleep_time_between_pages: int = 5,
        sort_by: Optional[str] = None,
        **kwargs,
    ):
        """."""
        self.http_conn_id = http_conn_id
        self.blade_get_token_op_id = blade_get_token_op_id
        self.sleep_time_between_pages = sleep_time_between_pages
        self.accepted_status_codes = accepted_status_codes
        self.sort_by = sort_by

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(op_id=self.http_conn_id)

        self.blade_get_token_op: BladeGetTokenOp = op_manager.op.get_op(
            op_id=self.blade_get_token_op_id
        ).op_obj

    def get_page(self, page: int) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """."""
        if self.sort_by:
            sort_string = f"&sorting={self.sort_by}"
        else:
            sort_string = ""

        # Example URL:
        # https://api.blade.co.uk/v1/products/variations
        # ?expand=*&&sorting=created&page=1
        url = (
            f"{self.http_conn.base_url}/v1/products/variations"
            f"?expand=*{sort_string}&page={page}"
        )
        logger.info(
            f"BladeProductsListVariationsOp.get_page: Fetching {page=} with {url=}"
        )

        etl_request_start_ts = pd.Timestamp.now(tz="UTC").isoformat()
        status_code, output_json = self.http_conn.call(
            method="get",
            url=url,
            requests_kwargs={
                "headers": {
                    "Content-Type": "application/json",
                    # Token gets refreshed if it has expired.
                    "Access-Token": self.blade_get_token_op.token,
                }
            },
            accepted_status_codes=self.accepted_status_codes,
            return_type="json",
        )
        etl_response_end_ts = pd.Timestamp.now(tz="UTC").isoformat()

        etl_metadata_json = {
            "etl_request_start_ts": etl_request_start_ts,
            "etl_response_end_ts": etl_response_end_ts,
        }

        return etl_metadata_json, output_json

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"BladeProductsListVariationsOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="BladeProductsListVariationsOp.run:",
        )

        page, total_pages, n_records_processed = 1, None, 0
        while (total_pages is None) or (page <= total_pages):

            if total_pages:
                logger.info(
                    f"BladeProductsListVariationsOp.run: Sleeping for "
                    f"{self.sleep_time_between_pages} seconds between pages. "
                    f"Next page is  {page=} of {total_pages=}."
                )
                time.sleep(self.sleep_time_between_pages)

            logger.info(
                f"BladeProductsListVariationsOp.run: "
                f"Fetching {page=} of {total_pages=}."
            )

            etl_metadata_json, output_json = self.get_page(page=page)
            n_records = len(output_json["data"])
            n_records_processed += n_records
            rate_limit_requests_remaining = output_json["meta"]["request_rate_limit"][
                "requests_remaining"
            ]

            # We get the total_pages from the data on every iteration. This is
            # because the total_pages can change as the data is being fetched.
            # NOTE: This may be flaky.
            total_pages = output_json["meta"]["paging"]["total_pages"]
            total_records = output_json["meta"]["paging"]["total_records"]

            logger.info(
                f"BladeProductsListVariationsOp.run: "
                f"Done fetching {page=} of {total_pages=}. "
                f"Fetched {n_records=} this request. Fetched "
                f"{n_records_processed}/{total_records} records so far. "
                f"{rate_limit_requests_remaining=}"
            )

            ############################################################################
            # NOTE: Adding `etl_metadata` into response json (output_json).
            # * This is to add some metadata about the ETL process, primarily some
            #   timestamps which may help with creating event stores/logs.
            # * This may also help with incremental modelling.
            # * DO NOT USE this information when calculating dedup_uuid.
            ############################################################################
            etl_metadata_json["time_step"] = time_step.to_json_dict()
            etl_metadata_json["blade"] = {"meta": output_json.get("meta", {})}
            output_json["etl_metadata"] = etl_metadata_json

            logger.info(
                f"BladeProductsListVariationsOp.run: Yielding results for page "
                f"{page}/{total_pages}."
            )

            yield OpMsg(
                data=output_json,
                metadata=BladeProductsListVariationsOpMetadataModel(
                    page=page,
                    total_pages=total_pages,
                    n_records=n_records,
                    n_records_processed=n_records_processed,
                    total_records=total_records,
                    rate_limit_requests_remaining=rate_limit_requests_remaining,
                ),
                audit=BladeProductsListVariationsOpAuditModel(),
            )

            # IMPORTANT: The page number is 1-indexed.
            # We increment the page number after we yield the data.
            page += 1



import hashlib
import logging
import uuid
from typing import Any, Dict, Generator, Optional

import bcrypt
from pydantic import StrictBool, StrictInt, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class BigcomGenericComputeEmailHashesOpConfigModel(BaseOpConfigModel):
    compute_hash_md5: StrictBool = True
    compute_hash_uuid5: StrictBool = True
    compute_hash_sha256: StrictBool = True
    compute_hash_sha512: StrictBool = True
    compute_hash_bcrypt: StrictBool = False
    bcrypt_salt_work_factor: StrictInt = 4

    @root_validator(pre=False)
    def validate_bcrypt_config(cls, values):
        # If compute_hash_bcrypt is True, bcrypt_salt_work_factor must be set.
        if values.get("compute_hash_bcrypt") is True and not isinstance(
            values.get("bcrypt_salt_work_factor", None), int
        ):
            raise ValueError(
                "If compute_hash_bcrypt is True, bcrypt_salt_work_factor must be set."
            )
        return values

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BigcomGenericComputeEmailHashesOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BigcomGenericComputeEmailHashesOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class BigcomGenericComputeEmailHashesOp(BaseGeneratorOp):
    """."""

    OP_TYPE = "generator.bigcom.generic.compute_email_hashes"
    OP_CONFIG_MODEL = BigcomGenericComputeEmailHashesOpConfigModel
    OP_METADATA_MODEL = BigcomGenericComputeEmailHashesOpMetadataModel
    OP_AUDIT_MODEL = BigcomGenericComputeEmailHashesOpAuditModel

    templated_fields = None

    def __init__(
        self,
        compute_hash_md5: bool = True,
        compute_hash_uuid5: bool = True,
        compute_hash_sha256: bool = True,
        compute_hash_sha512: bool = True,
        compute_hash_bcrypt: bool = False,
        bcrypt_salt_work_factor: int = 4,
        **kwargs,
    ):
        """."""
        self.compute_hash_md5 = compute_hash_md5
        self.compute_hash_uuid5 = compute_hash_uuid5
        self.compute_hash_sha256 = compute_hash_sha256
        self.compute_hash_sha512 = compute_hash_sha512
        self.compute_hash_bcrypt = compute_hash_bcrypt
        self.bcrypt_salt_work_factor = bcrypt_salt_work_factor
        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"BigcomGenericComputeEmailHashesOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="BigcomGenericComputeEmailHashesOp.run:",
        )

        input_records = msg.data

        if not isinstance(input_records, list):
            raise NotImplementedError("Input records must be a list of dicts.")

        final_records = []

        def compute_hashes_and_update_in_place(payload: Dict[str, Any], email_key: str):
            if self.compute_hash_md5:
                payload["email_address_hash_md5"] = hashlib.md5(
                    payload[email_key].encode("utf-8")
                ).hexdigest()

            if self.compute_hash_uuid5:
                payload["email_address_hash_uuid5"] = str(
                    uuid.uuid5(uuid.NAMESPACE_URL, payload[email_key])
                )

            if self.compute_hash_sha256:
                payload["email_address_hash_sha256"] = hashlib.sha256(
                    payload[email_key].encode("utf-8")
                ).hexdigest()

            if self.compute_hash_sha512:
                payload["email_address_hash_sha512"] = hashlib.sha512(
                    payload[email_key].encode("utf-8")
                ).hexdigest()

            if self.compute_hash_bcrypt:
                wf = self.bcrypt_salt_work_factor
                payload[f"email_address_hash_bcrypt_wf_{wf}"] = bcrypt.hashpw(
                    payload[email_key].encode("utf-8"),
                    bcrypt.gensalt(rounds=self.bcrypt_salt_work_factor),
                ).decode("utf-8")

        for record in input_records:
            email_exists = False

            if "email" in record["data"]:
                email_exists = True
                compute_hashes_and_update_in_place(
                    payload=record["data"], email_key="email"
                )

            if "billing_address" in record["data"]:
                if "email" in record["data"]["billing_address"]:
                    email_exists = True
                    compute_hashes_and_update_in_place(
                        payload=record["data"]["billing_address"], email_key="email"
                    )

            if not email_exists:
                raise ValueError(
                    f"BigcomGenericComputeEmailHashesOp.run: "
                    f"No email address found in record."
                )

            # record['data'] has been updated in place.
            final_records.append(record)

        yield OpMsg(
            data=final_records,
            metadata=BigcomGenericComputeEmailHashesOpMetadataModel(),
            audit=BigcomGenericComputeEmailHashesOpAuditModel(),
        )



import logging
import queue
from typing import Generator, Optional, Union

from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.queue import QueueConnOp

logger = logging.getLogger(__name__)


class QueueGetGeneratorOpConfigModel(BaseOpConfigModel):
    queue_conn_id: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class QueueGetGeneratorOpMetadataModel(BaseOpMetadataModel):
    item_i: int

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class QueueGetGeneratorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class QueueGetGeneratorOp(BaseGeneratorOp):
    OP_TYPE = "generator.queue.get"
    OP_CONFIG_MODEL = QueueGetGeneratorOpConfigModel
    OP_METADATA_MODEL = QueueGetGeneratorOpMetadataModel
    OP_AUDIT_MODEL = QueueGetGeneratorOpAuditModel

    templated_fields = None

    def __init__(
        self,
        queue_conn_id: str,
        **kwargs,
    ):
        self.queue_conn_id = queue_conn_id
        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.queue_conn: Union[QueueConnOp] = op_manager.get_connector(
            op_id=self.queue_conn_id
        )

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"QueueGetGeneratorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="QueueGetGeneratorOp.run:"
        )

        fifo: queue.Queue = self.queue_conn.get_reference()

        for item_i, item in enumerate(fifo.queue):
            logger.info(
                f"QueueGetGeneratorOp.run: Yielding {item_i=} (0-index) from queue of "
                f"type {type(item)=}"
            )
            yield OpMsg(
                data=item,
                metadata=QueueGetGeneratorOpMetadataModel(item_i=item_i),
                audit=QueueGetGeneratorOpAuditModel(),
            )



from typing import Any, Dict

from pydantic import BaseModel, Field, StrictBool, StrictStr, conint


def default_formatters() -> Dict[str, Any]:
    return {
        "default": {
            "format": (
                "[%(asctime)s][%(filename)-28s:%(lineno)-4d][%(levelname)9s] %(message)s"
            ),
        },
    }


def default_handlers() -> Dict[str, Any]:
    return {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
            "level": "DEBUG",
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "default",
            "level": "DEBUG",
            "filename": "logs/all.log",
            "maxBytes": 1024 * 1024,
            "backupCount": 100,
        },
    }


def default_loggers() -> Dict[str, Any]:
    # TODO: Better handing of root and nr_ops and airflow loggers.
    #  - currently, the file has duplicate logs for nr_ops and airflow
    return {
        "": {
            "handlers": ["file"],
            "level": "INFO",
            "propagate": False,
        },
        "airflow": {
            "handlers": ["console", "file"],
            "level": "WARNING",
            "propagate": True,
        },
        "nr_ops": {
            "handlers": ["console", "file"],
            "level": "DEBUG",
            "propagate": False,
        },
    }


class LoggingDictConfigModel(BaseModel):
    version: conint(ge=1) = 1
    disable_existing_loggers: StrictBool = False
    formatters: Dict[StrictStr, Dict[StrictStr, StrictStr]] = Field(
        default_factory=default_formatters
    )
    handlers: Dict[StrictStr, Dict[StrictStr, Any]] = Field(
        default_factory=default_handlers
    )
    loggers: Dict[StrictStr, Dict[StrictStr, Any]] = Field(
        default_factory=default_loggers
    )

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False



""".
https://googleapis.github.io/google-api-python-client/docs/dyn/analyticsreporting_v4.html
"""
import logging
from typing import Any, Dict, List, Literal, Optional

from google.analytics.data_v1beta.types import (
    DateRange,
    Dimension,
    Metric,
    RunReportRequest,
    RunReportResponse,
)
from pydantic import StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel
from nr_ops.ops.ops.connector_ops.interfaces.base import validate_hook_type_and_config

logger = logging.getLogger(__name__)


class GoogleAnalyticsGA4ConnOpConfigModel(BaseOpConfigModel):
    hook_type: Literal[
        "connector.hooks.airflow_gcp_hook",
        "connector.hooks.gcp_service_account_from_file",
        "connector.hooks.gcp_service_account_from_env",
    ]
    hook_config: Dict[StrictStr, Any]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False

    _validate_hook_type_and_config = root_validator(allow_reuse=True)(
        validate_hook_type_and_config
    )


class GoogleAnalyticsGA4ConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAnalyticsGA4ConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAnalyticsGA4ConnOp(BaseConnectorOp):
    OP_TYPE = "connector.google_analytics_ga4"
    OP_CONFIG_MODEL = GoogleAnalyticsGA4ConnOpConfigModel
    OP_METADATA_MODEL = GoogleAnalyticsGA4ConnOpMetadataModel
    OP_AUDIT_MODEL = GoogleAnalyticsGA4ConnOpAuditModel

    templated_fields = None

    def __init__(self, hook_type: str, hook_config: Dict[str, Any], **kwargs):
        self.hook_type = hook_type
        self.hook_config = hook_config
        self.templated_fields = kwargs.get("templated_fields", [])

        from nr_ops.ops.op_collection import OP_COLLECTION

        self.hook_op = OP_COLLECTION[self.hook_type](**self.hook_config)
        self.hook = self.hook_op.run().data

        # Will be poplated in run()
        self.analytics: Any = None

    def get_report(
        self,
        property_id: str,
        date_ranges: Optional[List[Dict[str, str]]] = None,
        metrics: Optional[List[Dict[str, str]]] = None,
        dimensions: Optional[List[Dict[str, str]]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """."""
        metric_names = []
        dimension_names = []
        if (
            self.hook_type == "connector.hooks.airflow_gcp_hook"
            or self.hook_type == "connector.hooks.gcp_service_account_from_file"
            or self.hook_type == "connector.hooks.gcp_service_account_from_env"
        ):
            payload = {"property": f"properties/{property_id}"}
            if date_ranges:
                payload["date_ranges"] = [
                    DateRange(**date_range) for date_range in date_ranges
                ]
            if metrics:
                payload["metrics"] = []
                for metric in metrics:
                    payload["metrics"].append(Metric(**metric))
                    metric_names.append(metric["name"])
            if dimensions:
                payload["dimensions"] = []
                for dimension in dimensions:
                    payload["dimensions"].append(Dimension(**dimension))
                    dimension_names.append(dimension["name"])

            from google.analytics.data_v1beta import BetaAnalyticsDataClient

            self.analytics: BetaAnalyticsDataClient
            request = RunReportRequest(**payload)
            results: RunReportResponse = self.analytics.run_report(request)
            data = [
                {
                    "metrics": {
                        metric_header.name: row.metric_values[i].value
                        for i, metric_header in enumerate(results.metric_headers)
                    },
                    "dimensions": {
                        dimension_header.name: row.dimension_values[i].value
                        for i, dimension_header in enumerate(results.dimension_headers)
                    },
                }
                for row in results.rows
            ]
            output = {
                "ga_type": "ga4",
                "request": {
                    "property_id": property_id,
                    "date_ranges": date_ranges,
                    "metrics": metrics,
                    "dimensions": dimensions,
                },
                "response": {
                    "property_id": property_id,
                    "date_ranges": date_ranges,
                    "data": data,
                },
            }
            return output
        else:
            raise NotImplementedError()

    def run(self) -> OpMsg:
        """."""
        logger.info(f"GoogleAnalyticsGA4ConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="GoogleAnalyticsGA4ConnOp.run:"
        )

        if self.hook_type == "connector.hooks.airflow_gcp_hook":
            # Type annotate for pycharm/mypy/autocomplete
            from airflow.providers.google.common.hooks.base_google import GoogleBaseHook

            self.hook: GoogleBaseHook

            import google.auth.credentials
            from google.analytics.data_v1beta import BetaAnalyticsDataClient

            credentials: google.auth.credentials.Credentials = (
                self.hook._get_credentials()
            )
            self.analytics = BetaAnalyticsDataClient(credentials=credentials)
        # Source: https://googleapis.github.io/google-api-python-client/docs/dyn/analyticsreporting_v4.html
        elif self.hook_type == "connector.hooks.gcp_service_account_from_file":
            # Type annotate for pycharm/mypy/autocomplete
            from google.analytics.data_v1beta import BetaAnalyticsDataClient
            from google.api_core.client_options import ClientOptions

            from nr_ops.ops.ops.connector_ops.hooks.gcp_service_account_from_file import (
                GCPServiceAccountFromFileConnOp,
            )

            self.hook: GCPServiceAccountFromFileConnOp
            client_options = ClientOptions(credentials_file=self.hook.key_file_path)
            self.analytics = BetaAnalyticsDataClient(client_options=client_options)

        elif self.hook_type == "connector.hooks.gcp_service_account_from_env":
            from google.analytics.data_v1beta import BetaAnalyticsDataClient
            from oauth2client.service_account import ServiceAccountCredentials

            credentials: ServiceAccountCredentials = self.hook.credentials

            self.analytics = BetaAnalyticsDataClient(credentials=credentials)

        else:
            raise NotImplementedError()

        return OpMsg(
            data=self,
            metadata=GoogleAnalyticsGA4ConnOpMetadataModel(),
            audit=GoogleAnalyticsGA4ConnOpAuditModel(),
        )
